from __future__ import annotations
import os
import attrs
import typing as ty
from collections import OrderedDict
from pathlib import Path
import logging
from copy import copy, deepcopy
import attrs.converters
import pydra.mark
from pydra.engine.core import Workflow
from arcana.core.exceptions import (
    ArcanaNameError,
    ArcanaUsageError,
    ArcanaDesignError,
    ArcanaPipelinesStackError,
    ArcanaOutputNotProducedException,
    ArcanaDataMatchError,
)
from fileformats.core.base import DataType, FileSet
from fileformats.field import Field
from fileformats.core.exceptions import FormatConversionError
import arcana.core.data.set.base
import arcana.core.data.row
from ..data.space import DataSpace
from ..utils.misc import (
    func_task,
    pydra_eq,
    path2varname,
    add_exc_note,
)
from ..utils.serialize import (
    asdict,
    fromdict,
    pydra_asdict,
    pydra_fromdict,
    ClassResolver,
    ObjectListConverter,
)


logger = logging.getLogger("arcana")


@attrs.define
class PipelineField:
    """Defines an input to a pipeline

    Parameters
    ----------
    name : str
        Name of the input and how it will be referred to in UI
    field : str, optional
        the name of the pydra input field to connect to, defaults to name
    datatype : type, optional
        the type of the items to be passed to the input, fileformats.generic.File by default
    """

    name: str
    field: str = attrs.field()
    datatype: type = attrs.field(
        default=None,
        converter=ClassResolver(
            DataType,
            allow_none=True,
            alternative_types=[arcana.core.data.row.DataRow],
        ),
    )

    @field.default
    def field_default(self):
        return self.name


logger = logging.getLogger("arcana")


@attrs.define
class Pipeline:
    """A thin wrapper around a Pydra workflow to link it to sources and sinks
    within a dataset

    Parameters
    ----------
    name : str
        the name of the pipeline, used to differentiate it from others
    row_frequency : DataSpace, optional
        The row_frequency of the pipeline, i.e. the row_frequency of the
        derivatvies within the dataset, e.g. per-session, per-subject, etc,
        by default None
    workflow : Workflow
        The pydra workflow that performs the actual analysis
    inputs : Sequence[ty.Union[str, ty.Tuple[str, type]]]
        List of column names (i.e. either data sources or sinks) to be
        connected to the inputs of the pipeline. If the pipelines requires
        the input to be in a datatype to the source, then it can be specified
        in a tuple (NAME, FORMAT)
    outputs : Sequence[ty.Union[str, ty.Tuple[str, type]]]
        List of sink names to be connected to the outputs of the pipeline
        If the input to be in a specific datatype, then it can be provided in
        a tuple (NAME, FORMAT)
    converter_args : dict[str, dict]
        keyword arguments passed on to the converter to control how the
        conversion is performed.
    dataset : Dataset
        the dataset the pipeline has been applied to
    """

    name: str = attrs.field()
    row_frequency: DataSpace = attrs.field()
    workflow: Workflow = attrs.field(eq=attrs.cmp_using(pydra_eq))
    inputs: list[PipelineField] = attrs.field(
        converter=ObjectListConverter(PipelineField)
    )
    outputs: list[PipelineField] = attrs.field(
        converter=ObjectListConverter(PipelineField)
    )
    converter_args: ty.Dict[str, dict] = attrs.field(
        factory=dict, converter=attrs.converters.default_if_none(factory=dict)
    )
    dataset: arcana.core.data.set.base.Dataset = attrs.field(
        metadata={"asdict": False}, default=None, eq=False, hash=False
    )

    def __attrs_post_init__(self):
        for field in self.inputs + self.outputs:
            if field.datatype is None:
                field.datatype = self.dataset[field.name].datatype

    @inputs.validator
    def inputs_validator(self, _, inputs: list[PipelineField]):
        for inpt in inputs:
            if inpt.datatype is arcana.core.data.row.DataRow:  # special case
                continue
            if self.dataset:
                column = self.dataset[inpt.name]
                # Check that a converter can be found if required
                if inpt.datatype:
                    try:
                        inpt.datatype.get_converter(column.datatype, name="dummy")
                    except FormatConversionError as e:
                        msg = (
                            f"required to in conversion of '{inpt.name}' input "
                            f"to '{self.name}' pipeline"
                        )
                        add_exc_note(e, msg)
                        raise
            elif inpt.datatype is None:
                raise ValueError(
                    f"Datatype must be explicitly set for {inpt.name} in unbound Pipeline"
                )
            if inpt.field not in self.workflow.input_names:
                raise ArcanaNameError(
                    inpt.field,
                    f"{inpt.field} is not in the input spec of '{self.name}' "
                    f"pipeline: " + "', '".join(self.workflow.input_names),
                )

    @outputs.validator
    def outputs_validator(self, _, outputs: list[PipelineField]):
        for outpt in outputs:
            if self.dataset:
                column = self.dataset[outpt.name]
                if column.row_frequency != self.row_frequency:
                    raise ArcanaUsageError(
                        f"Pipeline row_frequency ('{str(self.row_frequency)}') doesn't match "
                        f"that of '{outpt.name}' output ('{str(self.row_frequency)}')"
                    )
                # Check that a converter can be found if required
                if outpt.datatype:
                    try:
                        column.datatype.get_converter(outpt.datatype, name="dummy")
                    except FormatConversionError as e:
                        msg = (
                            f"required to in conversion of '{outpt.name}' output "
                            f"from '{self.name}' pipeline"
                        )
                        add_exc_note(e, msg)
                        raise
            elif outpt.datatype is None:
                raise ValueError(
                    f"Datatype must be explicitly set for {outpt.name} in unbound Pipeline"
                )
            if outpt.field not in self.workflow.output_names:
                raise ArcanaNameError(
                    f"{outpt.field} is not in the output spec of '{self.name}' "
                    f"pipeline: " + "', '".join(self.workflow.output_names)
                )

    @property
    def input_varnames(self):
        return [
            i.name for i in self.inputs
        ]  # [path2varname(i.name) for i in self.inputs]

    @property
    def output_varnames(self):
        return [
            o.name for o in self.outputs
        ]  # [path2varname(o.name) for o in self.outputs]

    # parameterisation = self.get_parameterisation(kwargs)
    # self.wf.to_process.inputs.parameterisation = parameterisation
    # self.wf.per_node.source.inputs.parameterisation = parameterisation

    def __call__(self, **kwargs):
        """
        Create an "outer" workflow that interacts with the dataset to pull input
        data, process it and then push the derivatives back to the store.

        Parameters
        ----------
        **kwargs
            passed directly to the Pydra.Workflow init. The `ids` arg can be
            used to filter the data rows over which the pipeline is run.

        Returns
        -------
        pydra.Workflow
            a Pydra workflow that iterates through the dataset, pulls data to the
            processing node, executes the analysis workflow on each data row,
            then uploads the outputs back to the data store

        Raises
        ------
        ArcanaUsageError
            If the new pipeline will overwrite an existing pipeline connection
            with overwrite == False.
        """

        # Create the outer workflow to link the analysis workflow with the
        # data row iteration and store connection rows
        wf = Workflow(name=self.name, input_spec=["ids"], **kwargs)

        # Generate list of rows to process checking existing outputs
        wf.add(
            to_process(
                dataset=self.dataset,
                row_frequency=self.row_frequency,
                outputs=self.outputs,
                requested_ids=None,  # FIXME: Needs to be set dynamically
                name="to_process",
            )
        )

        # Create the workflow that will be split across all rows for the
        # given data row_frequency
        wf.add(
            Workflow(
                name="per_row", input_spec=["id"], id=wf.to_process.lzout.ids
            ).split("id")
        )

        # Automatically output interface for source node to include sourced
        # columns
        source_out_dct = {}
        for inpt in self.inputs:
            # If the row frequency of the column is not a parent of the pipeline
            # then the input will be a sequence of all the child rows
            if inpt.datatype is arcana.core.data.row.DataRow:
                dtype = arcana.core.data.row.DataRow
            else:
                dtype = self.dataset[inpt.name].datatype
                # If the row frequency of the source column is higher than the frequency
                # of the pipeline, then the related elements of the source column are
                # collected into a list and passed to the pipeline
                if not self.dataset[inpt.name].row_frequency.is_parent(
                    self.row_frequency, if_match=True
                ):
                    dtype = list[dtype]
            source_out_dct[inpt.name] = dtype
        source_out_dct["provenance_"] = ty.Dict[str, ty.Any]

        wf.per_row.add(
            func_task(
                source_items,
                in_fields=[
                    ("dataset", arcana.core.data.set.base.Dataset),
                    ("row_frequency", DataSpace),
                    ("id", str),
                    ("inputs", ty.List[PipelineField]),
                    ("parameterisation", ty.Dict[str, ty.Any]),
                ],
                out_fields=list(source_out_dct.items()),
                name="source",
                dataset=self.dataset,
                row_frequency=self.row_frequency,
                inputs=self.inputs,
                id=wf.per_row.lzin.id,
            )
        )

        # Set the inputs
        sourced = {
            i.name: getattr(wf.per_row.source.lzout, i.name) for i in self.inputs
        }

        # Do input datatype conversions if required
        for inpt in self.inputs:
            if inpt.datatype == arcana.core.data.row.DataRow:
                continue
            stored_format = self.dataset[inpt.name].datatype
            converter = inpt.datatype.get_converter(
                stored_format,
                name=f"{inpt.name}_input_converter",
                **self.converter_args.get(inpt.name, {}),
            )
            if converter is not None:  # None if no conversion required
                logger.info(
                    "Adding implicit conversion for input '%s' " "from %s to %s",
                    inpt.name,
                    stored_format.mime_like,
                    inpt.datatype.mime_like,
                )
                converter.inputs.in_file = sourced.pop(inpt.name)
                if issubclass(source_out_dct[inpt.name], ty.Sequence):
                    # Iterate over all items in the sequence and convert them
                    # separately
                    converter.split("to_convert")
                # Insert converter
                wf.per_row.add(converter)
                # Map converter output to input_interface
                sourced[inpt.name] = converter.lzout.out_file

        # Create identity row to accept connections from user-defined rows
        # via `set_output` method
        wf.per_row.add(
            func_task(
                access_paths_and_values,
                in_fields=[
                    (
                        i.name,
                        ty.Union[DataType, arcana.core.data.row.DataRow, os.PathLike],
                    )
                    for i in self.inputs
                ],
                out_fields=[(i.name, ty.Any) for i in self.inputs],
                name="input_interface",
                **sourced,
            )
        )

        # Add the "inner" workflow of the pipeline that actually performs the
        # analysis/processing
        wf.per_row.add(deepcopy(self.workflow))
        # Make connections to "inner" workflow
        for inpt in self.inputs:
            setattr(
                getattr(wf.per_row, self.workflow.name).inputs,
                inpt.field,
                getattr(wf.per_row.input_interface.lzout, inpt.name),
            )

        # Creates a row to accept values from user-defined rows and
        # encapsulate them into DataTypes
        wf.per_row.add(
            func_task(
                encapsulate_paths_and_values,
                in_fields=[("outputs", ty.List[PipelineField])]
                + [(o.name, ty.Union[str, Path]) for o in self.outputs],
                out_fields=[(o.name, DataType) for o in self.outputs],
                name="output_interface",
                outputs=self.outputs,
                **{
                    o.name: getattr(
                        getattr(wf.per_row, self.workflow.name).lzout, o.field
                    )
                    for o in self.outputs
                },
            )
        )

        # Set datatype converters where required
        to_sink = {
            o: getattr(wf.per_row.output_interface.lzout, o)
            for o in self.output_varnames
        }

        # Do output datatype conversions if required
        for outpt in self.outputs:
            stored_format = self.dataset[outpt.name].datatype
            sink_name = path2varname(outpt.name)
            converter = stored_format.get_converter(
                outpt.datatype,
                name=f"{sink_name}_output_converter",
                **self.converter_args.get(outpt.name, {}),
            )
            if converter:
                logger.info(
                    "Adding implicit conversion for output '%s' " "from %s to %s",
                    outpt.name,
                    outpt.datatype.mime_like,
                    stored_format.mime_like,
                )
                # Insert converter
                converter.inputs.in_file = to_sink.pop(sink_name)
                wf.per_row.add(converter)
                # Map converter output to workflow output
                to_sink[sink_name] = converter.lzout.out_file

        # Can't use a decorated function as we need to allow for dynamic
        # arguments
        wf.per_row.add(
            func_task(
                sink_items,
                in_fields=(
                    [
                        ("dataset", arcana.core.data.set.base.Dataset),
                        ("row_frequency", DataSpace),
                        ("id", str),
                        ("provenance", ty.Dict[str, ty.Any]),
                    ]
                    + [
                        (s, ty.Union[DataType, str, bytes, os.PathLike])
                        for s in to_sink
                    ]
                ),
                out_fields=[("id", str)],
                name="sink",
                dataset=self.dataset,
                row_frequency=self.row_frequency,
                id=wf.per_row.lzin.id,
                provenance=wf.per_row.source.lzout.provenance_,
                **to_sink,
            )
        )

        wf.per_row.set_output([("id", wf.per_row.sink.lzout.id)])

        wf.set_output(
            [
                ("processed", wf.per_row.lzout.id),
                ("couldnt_process", wf.to_process.lzout.cant_process),
            ]
        )

        return wf

    PROVENANCE_VERSION = "1.0"
    WORKFLOW_NAME = "processing"

    def asdict(self, required_modules=None):
        dct = asdict(self, omit=["workflow"], required_modules=required_modules)
        dct["workflow"] = pydra_asdict(self.workflow, required_modules=required_modules)
        return dct

    @classmethod
    def fromdict(cls, dct, **kwargs):
        return fromdict(dct, workflow=pydra_fromdict(dct["workflow"]), **kwargs)

    @classmethod
    def stack(cls, *sinks):
        """Determines the pipelines stack, in order of execution,
        required to generate the specified sink columns.

        Parameters
        ----------
        sinks : Iterable[DataSink or str]
            the sink columns, or their names, that are to be generated

        Returns
        -------
        list[tuple[Pipeline, list[DataSink]]]
            stack of pipelines required to produce the specified data sinks,
            along with the sinks each stage needs to produce.

        Raises
        ------
        ArcanaDesignError
            when there are circular references in the pipelines stack
        """

        # Stack of pipelines to process in reverse order of required execution
        stack = OrderedDict()

        def push_pipeline_on_stack(sink, downstream: ty.Tuple[Pipeline] = None):
            """
            Push a pipeline onto the stack of pipelines to be processed,
            detecting common upstream pipelines and resolving them to a single
            pipeline

            Parameters
            ----------
            sink: DataSink
                the sink to push its deriving pipeline for
            downstream : tuple[Pipeline]
                The pipelines directly downstream of the pipeline to be added.
                Used to detect circular dependencies
            """
            if downstream is None:
                downstream = []
            if sink.pipeline_name is None:
                raise ArcanaDesignError(
                    f"{sink} hasn't been connected to a pipeline yet"
                )
            pipeline = sink.dataset.pipelines[sink.pipeline_name]
            if sink.name not in pipeline.output_varnames:
                raise ArcanaOutputNotProducedException(
                    f"{pipeline.name} does not produce {sink.name}"
                )
            # Check downstream piplines for circular dependencies
            downstream_pipelines = [p for p, _ in downstream]
            if pipeline in downstream_pipelines:
                recur_index = downstream_pipelines.index(pipeline)
                raise ArcanaDesignError(
                    f"{pipeline} cannot be a dependency of itself. Call-stack:\n"
                    + "\n".join(
                        "{} ({})".format(p, ", ".join(ro))
                        for p, ro in (
                            [[pipeline, sink.name]] + downstream[: (recur_index + 1)]
                        )
                    )
                )
            if pipeline.name in stack:
                # Pop pipeline from stack in order to add it to the end of the
                # stack and ensure it is run before all downstream pipelines
                prev_pipeline, to_produce = stack.pop(pipeline.name)
                assert pipeline is prev_pipeline
                # Combined required output to produce
                to_produce.append(sink)
            else:
                to_produce = []
            # Add the pipeline to the stack
            stack[pipeline.name] = pipeline, to_produce
            # Recursively add all the pipeline's prerequisite pipelines to the stack
            for inpt in pipeline.inputs:
                inpt_column = sink.dataset[inpt.name]
                if inpt_column.is_sink:
                    try:
                        push_pipeline_on_stack(
                            inpt_column,
                            downstream=[(pipeline, to_produce)] + downstream,
                        )
                    except ArcanaPipelinesStackError as e:
                        e.msg += (
                            "\nwhich are required as inputs to the '{}' "
                            "pipeline to produce '{}'".format(
                                pipeline.name, "', '".join(s.name for s in to_produce)
                            )
                        )
                        raise e

        # Add all pipelines
        for sink in sinks:
            push_pipeline_on_stack(sink)

        return reversed(stack.values())


def append_side_car_suffix(name, suffix):
    """Creates a new combined field name out of a basename and a side car"""
    return f"{name}__o__{suffix}"


def split_side_car_suffix(name):
    """Splits the basename from a side car sufix (as combined by `append_side_car_suffix`"""
    return name.split("__o__")


@pydra.mark.task
@pydra.mark.annotate({"return": {"ids": ty.List[str], "cant_process": ty.List[str]}})
def to_process(
    dataset: arcana.core.data.set.base.Dataset,
    row_frequency: DataSpace,
    outputs: ty.List[PipelineField],
    requested_ids: ty.Union[list[str], None],
    parameterisation: dict[str, ty.Any],
):
    if requested_ids is None:
        requested_ids = dataset.row_ids(row_frequency)
    ids = []
    cant_process = []
    for row in dataset.rows(row_frequency, ids=requested_ids):
        # TODO: Should check provenance of existing rows to see if it matches
        empty = [row.cell(o.name).is_empty for o in outputs]
        if all(empty):
            ids.append(row.id)
        elif any(empty):
            cant_process.append(row.id)
    logger.debug(
        "Found %s ids to process, and can't process %s due to partially present outputs",
        ids,
        cant_process,
    )
    return ids, cant_process


def source_items(
    dataset: arcana.core.data.set.base.Dataset,
    row_frequency: DataSpace,
    id: str,
    inputs: ty.List[PipelineField],
    parameterisation: dict,
):
    """Selects the items from the dataset corresponding to the input
    sources and retrieves them from the store to a cache on
    the host

    Parameters
    ----------
    dataset : Dataset
        the dataset to source the data from
    row_frequency : DataSpace
        the frequency of the row to source the data from
    id : str
        the ID of the row to source from
    parameterisation : dict
        provenance information... can't remember why this was used here...
    """
    logger.debug("Sourcing %s", inputs)
    provenance = copy(parameterisation)
    sourced = []
    row = dataset.row(row_frequency, id)
    with dataset.store.connection:
        missing_inputs = {}
        for inpt in inputs:
            # If the required datatype is of type DataRow then provide the whole
            # row to the pipeline input
            if inpt.datatype == arcana.core.data.row.DataRow:
                sourced.append(row)
                continue
            try:
                sourced.append(row[inpt.name])
            except ArcanaDataMatchError as e:
                missing_inputs[inpt.name] = str(e)
        if missing_inputs:
            raise ArcanaDataMatchError("\n\n" + "\n\n".join(missing_inputs.values()))
    return tuple(sourced) + (provenance,)


def sink_items(dataset, row_frequency, id, provenance, **to_sink):
    """Stores items generated by the pipeline back into the store

    Parameters
    ----------
    dataset : Dataset
        the dataset to source the data from
    row_frequency : DataSpace
        the frequency of the row to source the data from
    id : str
        the ID of the row to source from
    provenance : dict
        provenance information to be stored alongside the generated data
    **to_sink : dict[str, DataType]
        data items to be stored in the data store
    """
    logger.debug("Sinking %s", to_sink)
    row = dataset.row(row_frequency, id)
    with dataset.store.connection:
        for outpt_name, output in to_sink.items():
            row.cell(outpt_name).item = output
    return id


def access_paths_and_values(**data_items):
    """Copies files into the CWD renaming so the basenames match
    except for extensions"""
    logger.debug("Extracting paths/values from %s", data_items)
    values = []
    for name, item in data_items.items():
        if isinstance(item, FileSet):
            cpy = item.copy_to(Path.cwd() / name, symlink=True, make_dirs=True)
            values.append(cpy)
        elif isinstance(item, Field):
            values.append(item.value)
        else:
            values.append(item)
    return tuple(values) if len(values) > 1 else values[0]


def encapsulate_paths_and_values(outputs, **kwargs):
    """Copies files into the CWD renaming so the basenames match
    except for extensions"""
    logger.debug("Encapsulating %s into %s", kwargs, outputs)
    items = []
    for outpt in outputs:
        val = kwargs[outpt.name]
        items.append(outpt.datatype(val))
    if len(items) > 1:
        return tuple(items)
    elif items:
        return items[0]
    else:
        return None


# Provenance mismatch detection methods salvaged from data.provenance

# def mismatches(self, other, include=None, exclude=None):
#     """
#     Compares information stored within provenance objects with the
#     exception of version information to see if they match. Matches are
#     constrained to the name_paths passed to the 'include' kwarg, with the
#     exception of sub-name_paths passed to the 'exclude' kwarg

#     Parameters
#     ----------
#     other : Provenance
#         The provenance object to compare against
#     include : list[list[str]] | None
#         Paths in the provenance to include in the match. If None all are
#         incluced
#     exclude : list[list[str]] | None
#         Paths in the provenance to exclude from the match. In None all are
#         excluded
#     """
#     if include is not None:
#         include_res = [self._gen_prov_path_regex(p) for p in include]
#     if exclude is not None:
#         exclude_res = [self._gen_prov_path_regex(p) for p in exclude]
#     diff = DeepDiff(self._prov, other._prov, ignore_order=True)
#     # Create regular expressions for the include and exclude name_paths in
#     # the datatype that deepdiff uses for nested dictionary/lists

#     def include_change(change):
#         if include is None:
#             included = True
#         else:
#             included = any(rx.match(change) for rx in include_res)
#         if included and exclude is not None:
#             included = not any(rx.match(change) for rx in exclude_res)
#         return included

#     filtered_diff = {}
#     for change_type, changes in diff.items():
#         if isinstance(changes, dict):
#             filtered = dict((k, v) for k, v in changes.items()
#                             if include_change(k))
#         else:
#             filtered = [c for c in changes if include_change(c)]
#         if filtered:
#             filtered_diff[change_type] = filtered
#     return filtered_diff

# @classmethod
# def _gen_prov_path_regex(self, file_path):
#     if isinstance(file_path, str):
#         if file_path.startswith('/'):
#             file_path = file_path[1:]
#         regex = re.compile(r"root\['{}'\].*"
#                             .format(r"'\]\['".join(file_path.split('/'))))
#     elif not isinstance(file_path, re.Pattern):
#         raise ArcanaUsageError(
#             "Provenance in/exclude name_paths can either be name_path "
#             "strings or regexes, not '{}'".format(file_path))
#     return regex



from __future__ import annotations
import typing as ty
from pathlib import Path
from itertools import chain
import re
import logging
import shlex
import shutil
import attrs
import yaml
from urllib.parse import urlparse
from deepdiff import DeepDiff
from neurodocker.reproenv import DockerRenderer
from arcana.core import __version__
from arcana.core.utils.serialize import (
    ObjectConverter,
    ObjectListConverter,
    ClassResolver,
)
from ..command.base import ContainerCommand
from .base import ArcanaImage
from .components import ContainerAuthor, License, KnownIssue


logger = logging.getLogger("arcana")


@attrs.define(kw_only=True)
class App(ArcanaImage):
    """A container image that contains a command with specific inputs and outputs to run.

    Parameters
    ----------
    name : str
        name of the package/pipeline
    version : str
        version of the package/pipeline
    org : str
        the organisation the image will be tagged within
    base_image : BaseImage, optional
        the base image to build from
    packages : Packages, optional
        the package manager used to install system packages (should match OS on base image)
    registry : str, optional
        the container registry the image is to be installed at
    info_url : str
        the url of a documentation page describing the package
    authors : list[ContainerAuthor]
        list of authors of the package
    description : str
        single line description to be when referring to the pipeline in UIs
    command : ContainerCommand
        description of the command that is to be run within the image
    licenses : list[dict[str, str]], optional
        specification of licenses required by the commands in the container. Each dict
        should contain the 'name' of the license and the 'destination' it should be
        installed inside the container.
    build_iteration : str, optional
        version of the specification relative to the package version, i.e. if the package
        version hasn't been updated but the specification has been altered, the spec
        version should be updated (otherwise builds will fail). The spec version should
        reset to "0" if the package version is updated.
    long_description : str
        Multi-line description of the pipeline used in documentation
    known_issues : dict
        Any known issues with the pipeline. To be used in auto-doc generation
    loaded_from : Path
        the file the spec was loaded from, if applicable
    """

    IN_DOCKER_SPEC_PATH = "/arcana-spec.yaml"

    SUBPACKAGE = "deploy"

    info_url: str = attrs.field()
    authors: ty.List[ContainerAuthor] = attrs.field(
        converter=ObjectListConverter(ContainerAuthor),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    description: str
    command: ContainerCommand = attrs.field(converter=ObjectConverter(ContainerCommand))
    licenses: list[License] = attrs.field(
        factory=dict,
        converter=ObjectListConverter(License),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    known_issues: list[KnownIssue] = attrs.field(
        factory=list,
        converter=ObjectListConverter(KnownIssue),
        metadata={"serializer": ObjectListConverter.asdict},
    )
    long_description: str = ""
    loaded_from: Path = attrs.field(default=None, metadata={"asdict": False})
    arcana_version: str = __version__

    def __attrs_post_init__(self):

        # Set back-references to this image in the command spec
        self.command.image = self

    @info_url.validator
    def info_url_validator(self, _, info_url):
        parsed = urlparse(info_url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(
                f"Could not parse info url '{info_url}', please include URL scheme"
            )

    def add_entrypoint(self, dockerfile: DockerRenderer, build_dir: Path):

        command_line = (
            self.command.activate_conda_cmd() + "arcana deploy pipeline-entrypoint"
        )

        dockerfile.entrypoint(shlex.split(command_line))

    def construct_dockerfile(self, build_dir: Path, **kwargs) -> DockerRenderer:
        """Constructs a dockerfile that wraps a with dependencies

        Parameters
        ----------
        build_dir : Path
            Path to the directory the Dockerfile will be written into copy any local
            files to
        **kwargs
            Passed onto the ArcanaImage.construct_dockerfile() method

        Returns
        -------
        DockerRenderer
            Neurodocker Docker renderer to construct dockerfile from
        """

        dockerfile = super().construct_dockerfile(build_dir, **kwargs)

        self.install_licenses(
            dockerfile,
            build_dir,
        )

        self.insert_spec(dockerfile, build_dir)

        self.add_entrypoint(dockerfile, build_dir)

        return dockerfile

    def install_licenses(
        self,
        dockerfile: DockerRenderer,
        build_dir: Path,
    ):
        """Generate Neurodocker instructions to install licenses within the container
        image

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        build_dir : Path
            path to build dir
        """
        # Copy licenses into build directory
        license_build_dir = build_dir / "licenses"
        license_build_dir.mkdir()
        for lic in self.licenses:
            if lic.store_in_image:
                if lic.source:
                    build_path = license_build_dir / lic.name
                    shutil.copyfile(lic.source, build_path)
                    dockerfile.copy(
                        source=[str(build_path.relative_to(build_dir))],
                        destination=str(lic.destination),
                    )
                else:
                    logger.warning(
                        "License file for '%s' was not provided, will attempt to download "
                        "from '%s' dataset-level column or site-wide license dataset at "
                        "runtime",
                        lic.name,
                        lic.column_name(lic.name),
                    )

    def insert_spec(self, dockerfile: DockerRenderer, build_dir):
        """Generate Neurodocker instructions to save the specification inside the built
        image to be used when running the command and comparing against future builds

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        spec : dict
            the specification used to build the image
        build_dir : Path
            path to build dir
        """
        self.save(build_dir / "arcana-spec.yaml")
        dockerfile.copy(
            source=["./arcana-spec.yaml"], destination=self.IN_DOCKER_SPEC_PATH
        )

    def save(self, yml_path: Path):
        """Saves the specification to a YAML file that can be loaded again

        Parameters
        ----------
        yml_path : Path
            path to file to save the spec to
        """
        yml_dct = self.asdict()
        yml_dct["type"] = ClassResolver.tostr(self, strip_prefix=False)
        with open(yml_path, "w") as f:
            yaml.dump(yml_dct, f)

    @classmethod
    def load(
        cls,
        yml: ty.Union[Path, dict],
        root_dir: Path = None,
        license_paths: dict[str, Path] = None,
        licenses_to_download: set[str] = None,
        **kwargs,
    ):
        """Loads a deploy-build specification from a YAML file

        Parameters
        ----------
        yml : Path or dict
            path to the YAML file to load or loaded dictionary
        root_dir : Path, optional
            path to the root directory from which a tree of specs are being loaded from.
            The name of the root directory is taken to be the organisation the image
            belongs to, and all nested directories above the YAML file will be joined by
            '.' and prepended to the name of the loaded spec.
        license_paths : dict[str, Path], optional
            Licenses that are provided at build time to be included in the image.
        licenses_to_download : set[str], optional
            Licenses that are to be downloaded at runtime. If `license_paths` is not
            None (i.e. how to access required licenses are to be specified) then required
            licenses that are not in license_paths need to be explicitly listed in
            `licenses_to_download` otherwise an error is raised
        **kwargs
            additional keyword arguments that override/augment the values loaded from
            the spec file

        Returns
        -------
        Self
            The loaded spec object
        """

        if isinstance(yml, str):
            yml = Path(yml)
        if isinstance(yml, Path):
            yml_dict = cls._load_yaml(yml)
            if type(yml_dict) is not dict:
                raise ValueError(f"{yml!r} didn't contain a dict!")

            if "name" not in yml_dict:
                if root_dir is not None:
                    yml_dict["name"] = ".".join(
                        yml.relative_to(root_dir).parent.parts + (yml.stem,)
                    )
                else:
                    yml_dict["name"] = yml.stem

            if "org" not in yml_dict:
                if root_dir is not None:
                    yml_dict["org"] = root_dir.name
                else:
                    yml_dict["org"] = None

            yml_dict["loaded_from"] = yml.absolute()
        else:
            yml_dict = yml

        yml_dict.pop("type", None)  # Remove "type" from dict if present

        # Override/augment loaded values from spec
        yml_dict.update(kwargs)

        image = cls(**yml_dict)

        # Explicitly override directive in loaded spec to store license in the image

        if license_paths is not None:
            for lic in image.licenses:
                if lic.name in licenses_to_download:
                    lic.store_in_image = False
                if lic.store_in_image:
                    try:
                        lic.source = license_paths[lic.name]
                    except KeyError:
                        raise RuntimeError(
                            f"{lic.name} license has not been provided when it is "
                            "specified to be stored in the image"
                        )

        return image

    @classmethod
    def _load_yaml(cls, yaml_file: ty.Union[Path, str]):
        def yaml_join(loader, node):
            seq = loader.construct_sequence(node)
            return "".join([str(i) for i in seq])

        # Add special constructors to handle joins and concatenations within the YAML
        yaml.SafeLoader.add_constructor(tag="!join", constructor=yaml_join)
        with open(yaml_file, "r") as f:
            dct = yaml.load(f, Loader=yaml.SafeLoader)
        return dct

    @classmethod
    def load_tree(cls, root_dir: Path, **kwargs) -> list:
        """Walk the given directory structure and load all specs found within it

        Parameters
        ----------
        root_dir : Path
            path to the base directory
        """
        if root_dir.is_file():
            return [cls.load(root_dir, **kwargs)]
        specs = []
        for path in chain(root_dir.rglob("*.yml"), root_dir.rglob("*.yaml")):
            if not any(p.startswith(".") for p in path.parts):

                logging.info("Found container image specification file '%s'", path)
                specs.append(cls.load(path, root_dir=root_dir, **kwargs))

        return specs

    def autodoc(self, doc_dir, flatten: bool):
        header = {
            "title": self.name,
            "weight": 10,
        }

        if self.loaded_from:
            header["source_file"] = str(self.loaded_from)

        if flatten:
            out_dir = doc_dir
        else:
            assert isinstance(doc_dir, Path)

            out_dir = doc_dir.joinpath(*self.name.split(".")[:-1])

            assert doc_dir in out_dir.parents or out_dir == doc_dir

            out_dir.mkdir(parents=True, exist_ok=True)

        with open(f"{out_dir}/{self.name}.md", "w") as f:
            f.write("---\n")
            yaml.dump(header, f)
            f.write("\n---\n\n")

            f.write("## Package Info\n")
            tbl_info = MarkdownTable(f, "Key", "Value")
            tbl_info.write_row("Name", self.name)
            tbl_info.write_row("App version", self.version)
            tbl_info.write_row(
                "Build iteration", self.build_iteration if self.build_iteration else "0"
            )
            tbl_info.write_row("Base image", escaped_md(self.base_image.reference))
            tbl_info.write_row(
                "Maintainer", f"{self.authors[0].name} ({self.authors[0].email})"
            )
            tbl_info.write_row("Info URL", self.info_url)
            tbl_info.write_row("Short description", self.description)
            for known_issue in self.known_issues:
                tbl_info.write_row("Known issues", known_issue.url)

            short_desc = self.long_description or self.description
            f.write(f"\n{short_desc}\n\n")

            if self.licenses:
                f.write("### Required licenses\n")

                tbl_lic = MarkdownTable(f, "Name", "URL", "Description")
                for lic in self.licenses:
                    tbl_lic.write_row(
                        lic.name,
                        escaped_md(lic.info_url),
                        lic.description.strip(),
                    )

                f.write("\n")

            f.write("## Command\n")

            tbl_cmd = MarkdownTable(f, "Key", "Value")

            # if self.command.configuration is not None:
            #     config = self.command.configuration
            #     # configuration keys are variable depending on the workflow class
            tbl_cmd.write_row("Task", ClassResolver.tostr(self.command.task))
            freq_name = (
                self.command.row_frequency.name
                if not isinstance(self.command.row_frequency, str)
                else re.match(r".*\[(\w+)\]", self.command.row_frequency).group(1)
            )
            tbl_cmd.write_row("Operates on", freq_name)

            f.write("#### Inputs\n")
            tbl_inputs = MarkdownTable(
                f,
                "Name",
                "Required data-type",
                "Default column data-type",
                "Description",
            )
            if self.command.inputs is not None:
                for inpt in self.command.inputs:
                    tbl_inputs.write_row(
                        escaped_md(inpt.name),
                        self._data_format_html(inpt.datatype),
                        self._data_format_html(inpt.default_column.datatype),
                        inpt.help_string,
                    )
                f.write("\n")

            f.write("#### Outputs\n")
            tbl_outputs = MarkdownTable(
                f,
                "Name",
                "Required data-type",
                "Default column data-type",
                "Description",
            )
            if self.command.outputs is not None:
                for outpt in self.command.outputs:
                    tbl_outputs.write_row(
                        escaped_md(outpt.name),
                        self._data_format_html(outpt.datatype),
                        self._data_format_html(outpt.default_column.datatype),
                        outpt.help_string,
                    )
                f.write("\n")

            if self.command.parameters is not None:
                f.write("#### Parameters\n")
                tbl_params = MarkdownTable(f, "Name", "Data type", "Description")
                for param in self.command.parameters:
                    tbl_params.write_row(
                        escaped_md(param.name),
                        escaped_md(ClassResolver.tostr(param.datatype)),
                        param.help_string,
                    )
                f.write("\n")

    def compare_specs(self, other, check_version=True):
        """Compares two build specs against each other and returns the difference

        Parameters
        ----------
        s1 : dict
            first spec
        s2 : dict
            second spec
        check_version : bool
            check the arcana version used to generate the specs

        Returns
        -------
        DeepDiff
            the difference between the specs
        """

        sdict = self.asdict()
        odict = other.asdict()

        def prep(s):
            dct = {
                k: v
                for k, v in s.items()
                if (not k.startswith("_") and (v or isinstance(v, bool)))
            }
            if check_version:
                if "arcana_version" not in dct:
                    dct["arcana_version"] = __version__
            else:
                del dct["arcana_version"]
            return dct

        diff = DeepDiff(prep(sdict), prep(odict), ignore_order=True)
        return diff

    # @classmethod
    # def load_in_image(cls, spec_path: Path = IN_DOCKER_SPEC_PATH):
    #     yml_dct = cls._load_yaml(spec_path)
    #     klass = ClassResolver(cls)(yml_dct.pop("type"))
    #     return klass.load(yml_dct)

    @classmethod
    def _data_format_html(cls, datatype):

        datatype_str = datatype.mime_like if not isinstance(datatype, str) else datatype

        return (
            f'<span data-toggle="tooltip" data-placement="bottom" title="{datatype_str}" '
            f'aria-label="{datatype_str}">{datatype_str}</span>'
        )

    DOCKERFILE_README_TEMPLATE = """
        The following Docker image was generated by Arcana v{} to enable the
        commands to be run in the XNAT container service. See
        https://raw.githubusercontent.com/Australian-Imaging-Service/arcana/main/LICENSE
        for licence.

        {}

        """


class MarkdownTable:
    def __init__(self, f, *headers: str) -> None:
        self.headers = tuple(headers)

        self.f = f
        self._write_header()

    def _write_header(self):
        self.write_row(*self.headers)
        self.write_row(*("-" * len(x) for x in self.headers))

    def write_row(self, *cols: str):
        cols = list(cols)
        if len(cols) > len(self.headers):
            raise ValueError(
                f"More entries in row ({len(cols)} than columns ({len(self.headers)})"
            )

        # pad empty column entries if there's not enough
        cols += [""] * (len(self.headers) - len(cols))

        # TODO handle new lines in col
        self.f.write(
            "|" + "|".join(str(col).replace("|", "\\|") for col in cols) + "|\n"
        )


def escaped_md(value: str) -> str:
    if not value:
        return ""
    return f"`{value}`"



import zipfile
import tempfile
from pathlib import Path
from fileformats.text import Plain as Text
from fileformats.archive import Zip
from arcana.testing import TestDataSpace
from arcana.stdlib import DirTree
from conftest import TEST_DATASET_BLUEPRINTS
from arcana.testing.tasks import concatenate


def test_pipeline(work_dir):
    dataset = TEST_DATASET_BLUEPRINTS["concatenate_test"].make_dataset(
        DirTree(), work_dir / "dataset"
    )

    dataset.add_source("file1", Text)
    dataset.add_source("file2", Text)
    dataset.add_sink("deriv", Text)

    pipeline = dataset.apply_pipeline(
        name="test_pipeline",
        workflow=concatenate(duplicates=2, name="concatenate"),
        inputs=[("file1", "in_file1"), ("file2", "in_file2")],
        outputs=[("deriv", "out_file")],
        row_frequency=TestDataSpace.abcd,
    )

    IDS = ["a0b0c0d0", "a0b0c0d1"]

    workflow = pipeline(cache_dir=work_dir / "pipeline-cache")
    workflow(ids=IDS, plugin="serial")

    for item in dataset["deriv"]:
        with open(item.fspath) as f:
            contents = f.read()
        assert contents == "\n".join(["file1.txt", "file2.txt"] * 2)


def test_pipeline_with_implicit_conversion(work_dir):
    """Input files are converted from zip to Text, concatenated and then
    written back as zip files into the data store"""
    dataset = TEST_DATASET_BLUEPRINTS["concatenate_zip_test"].make_dataset(
        DirTree(), work_dir / "dataset"
    )

    dataset.add_source("file1", Zip[Text])
    dataset.add_source("file2", Zip[Text])
    dataset.add_sink("deriv", Zip[Text])

    pipeline = dataset.apply_pipeline(
        name="test_pipeline",
        workflow=concatenate(duplicates=2, name="concatenate"),
        inputs=[("file1", "in_file1", Text), ("file2", "in_file2", Text)],
        outputs=[("deriv", "out_file", Text)],
        row_frequency=TestDataSpace.abcd,
    )

    IDS = ["a0b0c0d0", "a0b0c0d1"]

    with dataset.tree:
        workflow = pipeline(cache_dir=work_dir / "pipeline-cache")
        workflow(ids=IDS, plugin="serial")

    for item in dataset["deriv"]:
        tmp_dir = Path(tempfile.mkdtemp())
        with zipfile.ZipFile(item.fspath) as zfile:
            zfile.extractall(path=tmp_dir)
        with open(tmp_dir / "out_file.txt") as f:
            contents = f.read()
        assert contents == "\n".join(["file1.zip", "file2.zip"] * 2)



from pathlib import Path
import tempfile
import attrs
import pytest
import pydra
from arcana.stdlib import Samples
from arcana.testing.tasks import (
    concatenate,
    concatenate_reverse,
    multiply_contents,
    contents_are_numeric,
    identity_file,
)
from arcana.core.analysis.mark import (
    analysis,
    pipeline,
    parameter,
    column,
    inherit,
    map_from,
    value_of,
    switch,
    is_provided,
    check,
    subanalysis,
)
from arcana.core.analysis.spec import Operation, ARCANA_SPEC
from fileformats.text import Plain as Text
from fileformats.archive import Zip
from arcana.stdlib import DirTree
from arcana.testing import MockRemote
from arcana.core.analysis.salience import (
    CheckStatus,
    ColumnSalience as cs,
    ParameterSalience as ps,
    CheckSalience as chs,
)
from arcana.core.exceptions import ArcanaDesignError


def get_contents(fpath):
    with open(fpath) as f:
        return f.read().splitlines()


@pytest.fixture
def source_dir():
    return Path(tempfile.mkdtemp())


@pytest.fixture
def sample_dir1(source_dir):
    smp_dir = source_dir / "sample1"
    smp_dir.mkdir()
    return smp_dir


@pytest.fixture
def sample_dir2(source_dir):
    smp_dir = source_dir / "sample2"
    smp_dir.mkdir()
    return smp_dir


@pytest.fixture
def test_file1(sample_dir1):
    fpath = sample_dir1 / "file1.txt"
    with open(fpath, "w") as f:
        f.write("file1")
    return fpath


@pytest.fixture
def test_file2(sample_dir1):
    fpath = sample_dir1 / "file2.txt"
    with open(fpath, "w") as f:
        f.write("file2")
    return fpath


@pytest.fixture
def test_file3(sample_dir1):
    fpath = sample_dir1 / "file3.txt"
    with open(fpath, "w") as f:
        f.write("file3")
    return fpath


@pytest.fixture
def test_numeric_file1(sample_dir2):
    file1_path = sample_dir2 / "numeric_file1.txt"
    with open(file1_path, "w") as f:
        f.write("1")
    return file1_path


@pytest.fixture
def test_numeric_file2(sample_dir2):
    file2_path = sample_dir2 / "file2.txt"
    with open(file2_path, "w") as f:
        f.write("2")
    return file2_path


@pytest.fixture
def test_dataset(source_dir, test_file1, test_file2, test_file3):
    dataset = DirTree().define_dataset(source_dir, space=Samples, hierarchy=["sample"])
    dataset.add_source("a_column", Text, "file1")
    dataset.add_source("another_column", Text, "file2")
    dataset.add_source("yet_another_column", Text, "file3")
    return dataset


@pytest.fixture
def test_partial_numeric_dataset(
    source_dir, test_file1, test_file2, test_numeric_file1, test_numeric_file2
):
    dataset = DirTree().define_dataset(source_dir, space=Samples, hierarchy=["sample"])
    dataset.add_source("a_column", Text, "file1")
    dataset.add_source("another_column", Text, "file2")
    return dataset


@pytest.fixture(scope="session")
def Concat():
    @analysis(Samples)
    class _Concat:

        file1: Zip[Text] = column("an arbitrary text file", salience=cs.primary)
        file2: Text = column("another arbitrary text file", salience=cs.primary)
        concatenated: Text = column("the output of concatenating file1 and file2")

        duplicates: int = parameter(
            "the number of times to duplicate the concatenation", default=1
        )

        @pipeline(concatenated)
        def concat_pipeline(self, wf, file1: Text, file2: Text, duplicates: int):

            wf.add(
                concatenate(
                    name="a_node", in_file1=file1, in_file2=file2, duplicates=duplicates
                )
            )

            return wf.a_node.lzout.out_file

    return _Concat


@pytest.fixture(scope="session")
def ExtendedConcat(Concat):
    @analysis(Samples)
    class _ExtendedConcat(Concat):

        # Sources
        concatenated = inherit()
        file3: Text = column("Another file to concatenate", salience=cs.primary)

        # Sinks
        doubly_concatenated: Text = column("The doubly concatenated file")

        # Parameters
        # Change the default 'duplicates' value for first concat and inherit it into the
        # namespace so it can also be used for the second concatenation pipeline too
        duplicates = inherit(default=2)

        @pipeline(doubly_concatenated)
        def doubly_concat_pipeline(
            self, wf, concatenated: Text, file3: Text, duplicates: int
        ):

            wf.add(
                concatenate(
                    name="concat",
                    in_file1=concatenated,
                    in_file2=file3,
                    duplicates=duplicates,
                )
            )

            return wf.concat.lzout.out_file

    return _ExtendedConcat


@pytest.fixture(scope="session")
def ConcatWithCheck(Concat):
    @analysis(Samples)
    class _ConcatWithCheck(Concat):

        concatenated = inherit()

        duplicates = inherit()

        @check(concatenated, salience=chs.recommended)
        def num_lines_check(self, wf, concatenated: Text, duplicates: int):
            """Checks the number of lines in the concatenated file to see whether they
            match what is expected for the number of duplicates specified"""

            @pydra.mark.task
            def num_lines_equals(in_file: Path, num_lines: int) -> CheckStatus:
                with open(in_file) as f:
                    contents = f.read()
                if len(contents.splitlines()) == num_lines:
                    status = CheckStatus.probable_pass
                else:
                    status = CheckStatus.failed
                return status

            wf.add(
                num_lines_equals(
                    in_file=concatenated,
                    num_lines=2 * duplicates,
                    name="num_lines_check",
                )
            )

            return wf.num_lines_check.lzout.out

    return _ConcatWithCheck


@pytest.fixture(scope="session")
def OverridenConcat(Concat):
    @analysis(Samples)
    class _OverridenConcat(Concat):

        file1 = inherit(ref=Concat.file1)
        file2 = inherit()
        concatenated = inherit()

        duplicates = inherit(default=2)
        order: str = parameter(
            "perform the concatenation in reverse order, i.e. file2 and then file1",
            choices=["forward", "reversed"],
            default="forward",
        )

        @pipeline(
            concatenated,
            condition=((value_of(order) == "reversed") & is_provided(file1)),
        )
        def reverse_concat_pipeline(
            self, wf, file1: Text, file2: Text, duplicates: int
        ) -> Text:

            wf.add(
                concatenate_reverse(
                    name="concat", in_file1=file1, in_file2=file2, duplicates=duplicates
                )
            )

            return wf.concat.lzout.out_file

    return _OverridenConcat


@pytest.fixture(scope="session")
def ConcatWithSwitch(Concat):
    @analysis(Samples)
    class _ConcatWithSwitch(Concat):

        file1 = inherit()
        file2 = inherit()
        concatenated = inherit()
        multiplied: Text = column("contents of the concatenated files are multiplied")

        multiplier: int = parameter(
            "the multiplier used to apply", salience=ps.required
        )

        @switch
        def inputs_are_numeric(self, wf, file1: Text, file2: Text):

            wf.add(contents_are_numeric(in_file=file1, name="check_file1"))

            wf.add(contents_are_numeric(in_file=file2, name="check_file2"))

            @pydra.mark.task
            def boolean_and(val1, val2) -> bool:
                return val1 and val2

            wf.add(
                boolean_and(
                    val1=wf.check_file1.lzout.out,
                    val2=wf.check_file2.lzout.out,
                    name="bool_and",
                )
            )

            return wf.bool_and.lzout.out

        @pipeline(multiplied, switch=inputs_are_numeric)
        def multiply_pipeline(self, wf, concatenated, multiplier):

            wf.add(
                multiply_contents(
                    name="multiply", in_file=concatenated, multiplier=multiplier
                )
            )

            return wf.multiply.lzout.out

    return _ConcatWithSwitch


@pytest.fixture(scope="session")
def ConcatWithSubanalyses(ExtendedConcat, ConcatWithSwitch):
    @analysis(Samples)
    class _ConcatWithSubanalyses:

        # Source columns mapped from "sub1" subanalysis so they can be shared across
        # both sub-analyses. Note that they could just as easily have been mapped from
        # "sub1" or recreated from scratch and mapped into both
        file1 = map_from("sub1", "file1")
        file2 = map_from("sub1", "file2")

        # Sink columns generated within the subanalyses mapped back out to the global
        # namespace so they can be mapped into the other subanalysis
        concatenated = map_from("sub2", "concatenated")
        concat_and_multiplied = map_from("sub2", "multiplied")

        # Link the duplicates parameter across both subanalyses so it is always the same
        # by mapping a global parameter into both subanalyses
        common_duplicates = map_from("sub1", "duplicates", default=5, salience=ps.check)

        # Additional parameters such as "multiplier" can be accessed within the subanalysis
        # class after the analysis class has been initialised using the 'sub2.multiplier'

        sub1: ExtendedConcat = subanalysis(
            "sub-analysis to add the 'doubly_concat' pipeline",
            # Feed the multiplied sink column from sub2 into the source column file3 of
            # the extended class
            concatenated=concatenated,  # Saves calculating concatenated twice in both sub-analyses
            file3=concat_and_multiplied,
        )
        sub2: ConcatWithSwitch = subanalysis(
            "sub-analysis to add the 'multiply' pipeline",
            file1=file1,
            file2=file2,
            # Use the concatenated generated by sub1 to avoid running it twice
            duplicates=common_duplicates,
        )

    return _ConcatWithSubanalyses


@pytest.fixture
def ConcatWithNestedSubanalyses(ConcatWithSubanalyses, ExtendedConcat):
    @analysis(Samples)
    class _ConcatWithNestedSubanalyses:

        file1 = map_from("basic_sub", "file1")
        file2 = map_from("outer_sub2", "file1")

        concatenated = map_from("basic_sub", "doubly_concatenated")

        duplicates = map_from("basic_sub", "duplicates", default=4)

        # multiplier1 = map_from("outer_sub1.sub1", "multiplier")  NotImplemented yet

        basic_sub: ExtendedConcat = subanalysis("basic subanalysis", file2=file2)
        outer_sub1: ConcatWithSubanalyses = subanalysis(
            "first outer subanalysis",
            file1=file1,
            file2=file2,
            concatenated=concatenated,
        )

        outer_sub2: ConcatWithSubanalyses = subanalysis(
            "second outer subanalysis",
            file2=file1,
            common_duplicates=duplicates,
        )

    return _ConcatWithNestedSubanalyses


def test_analysis_basic(Concat, test_file1, test_file2, test_dataset):

    analysis_spec = Concat.__spec__

    assert list(analysis_spec.parameter_names) == ["duplicates"]

    assert list(analysis_spec.column_names) == [
        "concatenated",
        "file1",
        "file2",
    ]

    assert list(analysis_spec.pipeline_names) == ["concat_pipeline"]
    assert list(analysis_spec.check_names) == []
    assert list(analysis_spec.switch_names) == []
    assert list(analysis_spec.subanalysis_names) == []

    duplicates = analysis_spec.parameter("duplicates")
    assert duplicates.type is int
    assert duplicates.default == 1
    assert duplicates.salience == ps.recommended
    assert duplicates.defined_in == (Concat,)

    file1 = analysis_spec.column_spec("file1")
    assert file1.type is Zip[Text]
    assert file1.row_frequency == Samples.sample
    assert file1.salience == cs.primary
    assert file1.defined_in == (Concat,)
    assert file1.mapped_from is None

    file2 = analysis_spec.column_spec("file2")
    assert file2.type is Text
    assert file2.row_frequency == Samples.sample
    assert file2.salience == cs.primary
    assert file2.defined_in == (Concat,)
    assert file2.mapped_from is None

    concatenated = analysis_spec.column_spec("concatenated")
    assert concatenated.type is Text
    assert concatenated.row_frequency == Samples.sample
    assert concatenated.salience == cs.supplementary
    assert concatenated.defined_in == (Concat,)
    assert concatenated.mapped_from is None

    concat_pipeline = analysis_spec.pipeline_builder("concat_pipeline")
    assert concat_pipeline.name == "concat_pipeline"
    assert concat_pipeline.parameters == ("duplicates",)
    assert concat_pipeline.inputs == ("file1", "file2")
    assert concat_pipeline.outputs == ("concatenated",)
    assert concat_pipeline.method is Concat.concat_pipeline
    assert concat_pipeline.condition is None
    assert concat_pipeline.switch is None

    # Initialise class
    analysis = Concat(
        dataset=test_dataset, file1="a_column", file2="another_column", duplicates=3
    )
    assert analysis.file1 == test_dataset["a_column"]
    assert analysis.file2 == test_dataset["another_column"]
    assert analysis.concatenated is None
    assert analysis.duplicates == 3
    wf = pydra.Workflow(
        name="test_analysis",
        input_spec=["file1", "file2"],
        file1=test_file1,
        file2=test_file2,
    )
    concatenated = analysis.concat_pipeline(
        wf, file1=wf.lzin.file1, file2=wf.lzin.file2, duplicates=analysis.duplicates
    )
    wf.set_output(("concatenated", concatenated))
    result = wf(plugin="serial")
    assert get_contents(result.output.concatenated) == [
        "file1",
        "file2",
        "file1",
        "file2",
        "file1",
        "file2",
    ]


def test_analysis_extended(
    Concat, ExtendedConcat, test_file1, test_file2, test_file3, test_dataset
):

    analysis_spec = ExtendedConcat.__spec__

    assert sorted(analysis_spec.parameter_names) == ["duplicates"]

    assert sorted(analysis_spec.column_names) == [
        "concatenated",
        "doubly_concatenated",
        "file1",
        "file2",
        "file3",
    ]

    assert list(analysis_spec.pipeline_names) == [
        "concat_pipeline",
        "doubly_concat_pipeline",
    ]
    assert list(analysis_spec.check_names) == []
    assert list(analysis_spec.switch_names) == []
    assert list(analysis_spec.subanalysis_names) == []

    duplicates = analysis_spec.parameter("duplicates")
    assert duplicates.type is int
    assert duplicates.default == 2
    assert duplicates.salience == ps.recommended
    assert duplicates.defined_in == (Concat, ExtendedConcat)
    assert duplicates.modified == ((("default", 2),),)

    file1 = analysis_spec.column_spec("file1")
    assert file1.type is Zip[Text]
    assert file1.row_frequency == Samples.sample
    assert file1.salience == cs.primary
    assert file1.defined_in == (Concat,)
    assert file1.mapped_from is None

    file2 = analysis_spec.column_spec("file2")
    assert file2.type is Text
    assert file2.row_frequency == Samples.sample
    assert file2.salience == cs.primary
    assert file2.defined_in == (Concat,)
    assert file2.mapped_from is None

    file3 = analysis_spec.column_spec("file3")
    assert file3.type is Text
    assert file3.row_frequency == Samples.sample
    assert file3.salience == cs.primary
    assert file3.defined_in == (ExtendedConcat,)
    assert file3.mapped_from is None

    concatenated = analysis_spec.column_spec("concatenated")
    assert concatenated.type is Text
    assert concatenated.row_frequency == Samples.sample
    assert concatenated.salience == cs.supplementary
    assert concatenated.defined_in == (Concat,)
    assert concatenated.mapped_from is None

    doubly_concatenated = analysis_spec.column_spec("doubly_concatenated")
    assert doubly_concatenated.type is Text
    assert doubly_concatenated.row_frequency == Samples.sample
    assert doubly_concatenated.salience == cs.supplementary
    assert doubly_concatenated.defined_in == (ExtendedConcat,)

    concat_pipeline = analysis_spec.pipeline_builder("concat_pipeline")
    assert concat_pipeline.name == "concat_pipeline"
    assert concat_pipeline.parameters == ("duplicates",)
    assert concat_pipeline.inputs == ("file1", "file2")
    assert concat_pipeline.outputs == ("concatenated",)
    assert concat_pipeline.method is Concat.concat_pipeline
    assert concat_pipeline.defined_in == (Concat,)
    assert concat_pipeline.condition is None
    assert concat_pipeline.switch is None

    doubly_concat_pipeline = analysis_spec.pipeline_builder("doubly_concat_pipeline")
    assert doubly_concat_pipeline.name == "doubly_concat_pipeline"
    assert doubly_concat_pipeline.parameters == ("duplicates",)
    assert doubly_concat_pipeline.inputs == ("concatenated", "file3")
    assert doubly_concat_pipeline.outputs == ("doubly_concatenated",)
    assert doubly_concat_pipeline.method is ExtendedConcat.doubly_concat_pipeline
    assert doubly_concat_pipeline.defined_in == (ExtendedConcat,)
    assert doubly_concat_pipeline.condition is None
    assert doubly_concat_pipeline.switch is None

    # Initialise class
    analysis = ExtendedConcat(
        dataset=test_dataset,
        file1="a_column",
        file2="another_column",
        file3="yet_another_column",
        duplicates=1,
    )
    assert analysis.file1 == test_dataset["a_column"]
    assert analysis.file2 == test_dataset["another_column"]
    assert analysis.file3 == test_dataset["yet_another_column"]
    assert analysis.concatenated is None
    assert analysis.doubly_concatenated is None
    assert analysis.duplicates == 1
    wf = pydra.Workflow(
        name="test_analysis",
        input_spec=["file1", "file2", "file3"],
        file1=test_file1,
        file2=test_file2,
        file3=test_file3,
    )
    concatenated = analysis.concat_pipeline(
        wf, file1=wf.lzin.file1, file2=wf.lzin.file2, duplicates=analysis.duplicates
    )
    doubly_concatenated = analysis.doubly_concat_pipeline(
        wf,
        concatenated=concatenated,
        file3=wf.lzin.file3,
        duplicates=analysis.duplicates,
    )
    wf.set_output(("doubly_concatenated", doubly_concatenated))
    result = wf(plugin="serial")
    assert get_contents(result.output.doubly_concatenated) == [
        "file1",
        "file2",
        "file3",
    ]


def test_analysis_with_check(
    Concat, ConcatWithCheck, test_file1, test_file2, test_dataset
):

    analysis_spec = ConcatWithCheck.__spec__

    assert sorted(analysis_spec.parameter_names) == ["duplicates"]

    assert sorted(analysis_spec.column_names) == [
        "concatenated",
        "file1",
        "file2",
    ]

    assert list(analysis_spec.pipeline_names) == [
        "concat_pipeline",
    ]
    assert list(analysis_spec.check_names) == ["num_lines_check"]
    assert list(analysis_spec.switch_names) == []
    assert list(analysis_spec.subanalysis_names) == []

    duplicates = analysis_spec.parameter("duplicates")
    assert duplicates.type is int
    assert duplicates.default == 1
    assert duplicates.salience == ps.recommended
    assert duplicates.defined_in == (Concat,)

    file1 = analysis_spec.column_spec("file1")
    assert file1.type is Zip[Text]
    assert file1.row_frequency == Samples.sample
    assert file1.salience == cs.primary
    assert file1.defined_in == (Concat,)
    assert file1.mapped_from is None

    file2 = analysis_spec.column_spec("file2")
    assert file2.type is Text
    assert file2.row_frequency == Samples.sample
    assert file2.salience == cs.primary
    assert file2.defined_in == (Concat,)
    assert file2.mapped_from is None

    concatenated = analysis_spec.column_spec("concatenated")
    assert concatenated.type is Text
    assert concatenated.row_frequency == Samples.sample
    assert concatenated.salience == cs.supplementary
    assert concatenated.defined_in == (Concat,)
    assert concatenated.mapped_from is None

    concat_pipeline = analysis_spec.pipeline_builder("concat_pipeline")
    assert concat_pipeline.name == "concat_pipeline"
    assert concat_pipeline.parameters == ("duplicates",)
    assert concat_pipeline.inputs == ("file1", "file2")
    assert concat_pipeline.outputs == ("concatenated",)
    assert concat_pipeline.method is Concat.concat_pipeline
    assert concat_pipeline.condition is None
    assert concat_pipeline.switch is None
    assert concat_pipeline.defined_in == (Concat,)

    num_lines_check = analysis_spec.check("num_lines_check")
    assert num_lines_check.name == "num_lines_check"
    assert num_lines_check.column == "concatenated"
    assert num_lines_check.inputs == ("concatenated",)
    assert num_lines_check.parameters == ("duplicates",)
    assert num_lines_check.method == ConcatWithCheck.num_lines_check
    assert num_lines_check.column == "concatenated"
    assert num_lines_check.defined_in == (ConcatWithCheck,)

    # Initialise class
    analysis = ConcatWithCheck(
        dataset=test_dataset, file1="a_column", file2="another_column", duplicates=7
    )
    assert analysis.file1 == test_dataset["a_column"]
    assert analysis.file2 == test_dataset["another_column"]
    assert analysis.concatenated is None
    assert analysis.duplicates == 7
    wf = pydra.Workflow(
        name="test_analysis",
        input_spec=["file1", "file2"],
        file1=test_file1,
        file2=test_file2,
    )
    concatenated = analysis.concat_pipeline(
        wf, file1=wf.lzin.file1, file2=wf.lzin.file2, duplicates=analysis.duplicates
    )
    num_lines_check = analysis.num_lines_check(
        wf, concatenated=concatenated, duplicates=analysis.duplicates
    )
    wf.set_output(("num_lines_check", num_lines_check))
    result = wf(plugin="serial")
    assert result.output.num_lines_check == CheckStatus.probable_pass


def test_analysis_override(
    Concat, OverridenConcat, test_file1, test_file2, test_dataset
):
    """Tests overriding methods in the base class with optional switches based on
    parameters and properties of the inputs"""

    analysis_spec = OverridenConcat.__spec__

    assert list(analysis_spec.column_names) == [
        "concatenated",
        "file1",
        "file2",
    ]
    assert list(analysis_spec.parameter_names) == [
        "duplicates",
        "order",
    ]

    assert list(analysis_spec.pipeline_names) == [
        "concat_pipeline",
        "reverse_concat_pipeline",
    ]
    assert list(analysis_spec.check_names) == []
    assert list(analysis_spec.switch_names) == []
    assert list(analysis_spec.subanalysis_names) == []

    file1 = analysis_spec.column_spec("file1")
    assert file1.type is Zip[Text]
    assert file1.row_frequency == Samples.sample
    assert file1.salience == cs.primary
    assert file1.defined_in == (Concat,)
    assert file1.mapped_from is None

    file2 = analysis_spec.column_spec("file2")
    assert file2.type is Text
    assert file2.row_frequency == Samples.sample
    assert file2.salience == cs.primary
    assert file2.defined_in == (Concat,)
    assert file2.mapped_from is None

    concatenated = analysis_spec.column_spec("concatenated")
    assert concatenated.type is Text
    assert concatenated.row_frequency == Samples.sample
    assert concatenated.salience == cs.supplementary
    assert concatenated.defined_in == (Concat,)
    assert concatenated.mapped_from is None

    duplicates = analysis_spec.parameter("duplicates")
    assert duplicates.type is int
    assert duplicates.default == 2
    assert duplicates.salience == ps.recommended
    assert duplicates.defined_in == (Concat, OverridenConcat)

    order = analysis_spec.parameter("order")
    assert order.type is str
    assert order.default == "forward"
    assert order.salience == ps.recommended
    assert order.defined_in == (OverridenConcat,)

    concat_pipeline = analysis_spec.pipeline_builder("concat_pipeline")
    assert concat_pipeline.name == "concat_pipeline"
    assert concat_pipeline.parameters == ("duplicates",)
    assert concat_pipeline.inputs == ("file1", "file2")
    assert concat_pipeline.outputs == ("concatenated",)
    assert concat_pipeline.method is Concat.concat_pipeline
    assert concat_pipeline.defined_in == (Concat,)
    assert concat_pipeline.condition is None
    assert concat_pipeline.switch is None

    reverse_concat_pipeline = analysis_spec.pipeline_builder("reverse_concat_pipeline")
    assert reverse_concat_pipeline.name == "reverse_concat_pipeline"
    assert reverse_concat_pipeline.parameters == ("duplicates",)
    assert reverse_concat_pipeline.inputs == ("file1", "file2")
    assert reverse_concat_pipeline.outputs == ("concatenated",)
    assert reverse_concat_pipeline.method is OverridenConcat.reverse_concat_pipeline
    assert reverse_concat_pipeline.defined_in == (OverridenConcat,)
    assert isinstance(reverse_concat_pipeline.condition, Operation)
    assert reverse_concat_pipeline.switch is None

    # Initialise class
    analysis = OverridenConcat(
        dataset=test_dataset,
        file1="a_column",
        file2="another_column",
        duplicates=1,
        order="reversed",
    )
    assert analysis.file1 == test_dataset["a_column"]
    assert analysis.file2 == test_dataset["another_column"]
    assert analysis.concatenated is None
    assert analysis.duplicates == 1
    assert analysis.order == "reversed"
    wf = pydra.Workflow(
        name="test_analysis",
        input_spec=["file1", "file2"],
        file1=test_file1,
        file2=test_file2,
    )
    concatenated = analysis.reverse_concat_pipeline(
        wf, file1=wf.lzin.file1, file2=wf.lzin.file2, duplicates=analysis.duplicates
    )
    wf.set_output(("concatenated", concatenated))
    result = wf(plugin="serial")
    assert get_contents(result.output.concatenated) == ["1elif", "2elif"]


def test_analysis_switch(
    Concat,
    ConcatWithSwitch,
    test_numeric_file1,
    test_numeric_file2,
    test_partial_numeric_dataset,
):
    """Tests overriding methods in the base class with optional switches based on
    parameters and properties of the inputs"""

    analysis_spec = ConcatWithSwitch.__spec__

    assert list(analysis_spec.column_names) == [
        "concatenated",
        "file1",
        "file2",
        "multiplied",
    ]
    assert list(analysis_spec.parameter_names) == [
        "duplicates",
        "multiplier",
    ]

    assert list(analysis_spec.pipeline_names) == [
        "concat_pipeline",
        "multiply_pipeline",
    ]
    assert list(analysis_spec.check_names) == []
    assert list(analysis_spec.switch_names) == ["inputs_are_numeric"]
    assert list(analysis_spec.subanalysis_names) == []

    file1 = analysis_spec.column_spec("file1")
    assert file1.type is Zip[Text]
    assert file1.row_frequency == Samples.sample
    assert file1.salience == cs.primary
    assert file1.defined_in == (Concat,)
    assert file1.mapped_from is None

    file2 = analysis_spec.column_spec("file2")
    assert file2.type is Text
    assert file2.row_frequency == Samples.sample
    assert file2.salience == cs.primary
    assert file2.defined_in == (Concat,)
    assert file2.mapped_from is None

    concatenated = analysis_spec.column_spec("concatenated")
    assert concatenated.type is Text
    assert concatenated.row_frequency == Samples.sample
    assert concatenated.salience == cs.supplementary
    assert concatenated.defined_in == (Concat,)
    assert concatenated.mapped_from is None

    multiplied = analysis_spec.column_spec("multiplied")
    assert multiplied.type is Text
    assert multiplied.row_frequency == Samples.sample
    assert multiplied.salience == cs.supplementary
    assert multiplied.defined_in == (ConcatWithSwitch,)
    assert multiplied.mapped_from is None

    duplicates = analysis_spec.parameter("duplicates")
    assert duplicates.type is int
    assert duplicates.default == 1
    assert duplicates.salience == ps.recommended
    assert duplicates.defined_in == (Concat,)

    multiplier = analysis_spec.parameter("multiplier")
    assert multiplier.type is int
    assert multiplier.default is None
    assert multiplier.salience == ps.required
    assert multiplier.defined_in == (ConcatWithSwitch,)

    concat_pipeline = analysis_spec.pipeline_builder("concat_pipeline")
    assert concat_pipeline.name == "concat_pipeline"
    assert concat_pipeline.parameters == ("duplicates",)
    assert concat_pipeline.inputs == ("file1", "file2")
    assert concat_pipeline.outputs == ("concatenated",)
    assert concat_pipeline.method is Concat.concat_pipeline
    assert concat_pipeline.defined_in == (Concat,)
    assert concat_pipeline.condition is None
    assert concat_pipeline.switch is None

    multiply_pipeline = analysis_spec.pipeline_builder("multiply_pipeline")
    assert multiply_pipeline.name == "multiply_pipeline"
    assert multiply_pipeline.parameters == ("multiplier",)
    assert multiply_pipeline.inputs == ("concatenated",)
    assert multiply_pipeline.outputs == ("multiplied",)
    assert multiply_pipeline.method is ConcatWithSwitch.multiply_pipeline
    assert multiply_pipeline.defined_in == (ConcatWithSwitch,)
    assert multiply_pipeline.condition is None
    assert multiply_pipeline.switch == "inputs_are_numeric"

    inputs_are_numeric = analysis_spec.switch("inputs_are_numeric")
    assert inputs_are_numeric.name == "inputs_are_numeric"
    assert inputs_are_numeric.parameters == ()
    assert inputs_are_numeric.inputs == ("file1", "file2")
    assert inputs_are_numeric.method is ConcatWithSwitch.inputs_are_numeric
    assert inputs_are_numeric.defined_in == (ConcatWithSwitch,)

    # Initialise class
    analysis = ConcatWithSwitch(
        dataset=test_partial_numeric_dataset,
        file1="a_column",
        file2="another_column",
        duplicates=1,
        multiplier=10,
    )
    assert analysis.file1 == test_partial_numeric_dataset["a_column"]
    assert analysis.file2 == test_partial_numeric_dataset["another_column"]
    assert analysis.concatenated is None
    assert analysis.duplicates == 1
    assert analysis.multiplier == 10

    wf = pydra.Workflow(
        name="test_analysis",
        input_spec=["file1", "file2"],
        file1=test_numeric_file1,
        file2=test_numeric_file2,
    )
    inputs_are_numeric = analysis.inputs_are_numeric(
        wf, file1=wf.lzin.file1, file2=wf.lzin.file2
    )
    concatenated = analysis.concat_pipeline(
        wf, file1=wf.lzin.file1, file2=wf.lzin.file2, duplicates=analysis.duplicates
    )
    multiplied = analysis.multiply_pipeline(
        wf, concatenated=concatenated, multiplier=analysis.multiplier
    )
    wf.set_output(
        [("inputs_are_numeric", inputs_are_numeric), ("multiplied", multiplied)]
    )
    result = wf(plugin="serial")
    assert result.output.inputs_are_numeric is True
    assert get_contents(result.output.multiplied) == ["10.0", "20.0"]


def test_analysis_with_subanalyses(
    ConcatWithSubanalyses,
    ExtendedConcat,
    ConcatWithSwitch,
    test_numeric_file1,
    test_numeric_file2,
    test_partial_numeric_dataset,
):

    analysis_spec = ConcatWithSubanalyses.__spec__

    assert list(analysis_spec.parameter_names) == ["common_duplicates"]

    assert list(analysis_spec.column_names) == [
        "concat_and_multiplied",
        "concatenated",
        "file1",
        "file2",
    ]

    assert list(analysis_spec.pipeline_names) == []
    assert list(analysis_spec.check_names) == []
    assert list(analysis_spec.switch_names) == []
    assert list(analysis_spec.subanalysis_names) == ["sub1", "sub2"]

    common_duplicates = analysis_spec.parameter("common_duplicates")
    assert common_duplicates.type is int
    assert common_duplicates.default == 5
    assert common_duplicates.salience == ps.check
    # Not sure why this is failing, not super critical at this point
    # assert common_duplicates.defined_in == (ConcatWithSubanalyses,)

    file1 = analysis_spec.column_spec("file1")
    assert file1.type is Zip[Text]
    assert file1.row_frequency == Samples.sample
    assert file1.salience == cs.primary
    # assert file1.defined_in == (Concat,)
    assert file1.mapped_from == ("sub1", "file1")

    file2 = analysis_spec.column_spec("file2")
    assert file2.type is Text
    assert file2.row_frequency == Samples.sample
    assert file2.salience == cs.primary
    # assert file2.defined_in == (Concat,)
    assert file2.mapped_from == ("sub1", "file2")

    concat_and_multiplied = analysis_spec.column_spec("concat_and_multiplied")
    assert concat_and_multiplied.type is Text
    assert concat_and_multiplied.row_frequency == Samples.sample
    assert concat_and_multiplied.salience == cs.supplementary
    # assert concat_and_multiplied.defined_in == (ConcatWithSwitch,)
    assert concat_and_multiplied.mapped_from == ("sub2", "multiplied")

    sub1 = analysis_spec.subanalysis_spec("sub1")
    assert sub1.name == "sub1"
    assert sub1.type is ExtendedConcat
    assert sub1.mappings == (
        ("concatenated", "concatenated"),
        ("duplicates", "common_duplicates"),
        ("file1", "file1"),
        ("file2", "file2"),
        ("file3", "concat_and_multiplied"),
    )
    assert sub1.defined_in == (ConcatWithSubanalyses,)

    sub2 = analysis_spec.subanalysis_spec("sub2")
    assert sub2.name == "sub2"
    assert sub2.type is ConcatWithSwitch
    assert sub2.mappings == (
        ("concatenated", "concatenated"),
        ("duplicates", "common_duplicates"),
        ("file1", "file1"),
        ("file2", "file2"),
        ("multiplied", "concat_and_multiplied"),
    )
    assert sub2.defined_in == (ConcatWithSubanalyses,)

    # Initialise class
    analysis = ConcatWithSubanalyses(
        dataset=test_partial_numeric_dataset,
        file1="a_column",
        file2="another_column",
        common_duplicates=1,
        sub2={"multiplier": 100},
    )
    # analysis.sub2.multiplier = 100
    assert analysis.file1 == test_partial_numeric_dataset["a_column"]
    assert analysis.file2 == test_partial_numeric_dataset["another_column"]
    assert analysis.common_duplicates == 1
    assert analysis.sub1.file1 == test_partial_numeric_dataset["a_column"]
    assert analysis.sub1.file2 == test_partial_numeric_dataset["another_column"]
    assert analysis.sub2.duplicates == 1
    assert analysis.sub1.concatenated is None
    assert analysis.sub1.doubly_concatenated is None
    assert analysis.sub2.file1 == test_partial_numeric_dataset["a_column"]
    assert analysis.sub2.file2 == test_partial_numeric_dataset["another_column"]
    assert analysis.sub2.duplicates == 1
    assert analysis.sub2.multiplier == 100
    assert analysis.sub2.concatenated is None
    assert analysis.sub2.multiplied is None

    wf = pydra.Workflow(
        name="test_analysis",
        input_spec=["file1", "file2"],
        file1=test_numeric_file1,
        file2=test_numeric_file2,
    )
    concatenated = analysis.sub2.concat_pipeline(
        wf,
        file1=wf.lzin.file1,
        file2=wf.lzin.file2,
        duplicates=analysis.sub2.duplicates,
    )
    concat_and_multiplied = analysis.sub2.multiply_pipeline(
        wf, concatenated=concatenated, multiplier=analysis.sub2.multiplier
    )
    doubly_concatenated = analysis.sub1.doubly_concat_pipeline(
        wf,
        concatenated=concatenated,
        file3=concat_and_multiplied,
        duplicates=analysis.common_duplicates,
    )
    wf.set_output(("doubly_concatenated", doubly_concatenated))
    result = wf(plugin="serial")
    assert get_contents(result.output.doubly_concatenated) == [
        "1",
        "2",
        "100.0",
        "200.0",
    ]


def test_analysis_with_nested_subanalyses(
    ConcatWithNestedSubanalyses,
    ConcatWithSubanalyses,
    ExtendedConcat,
    test_numeric_file1,
    test_numeric_file2,
    test_partial_numeric_dataset,
):

    analysis_spec = ConcatWithNestedSubanalyses.__spec__

    assert sorted(analysis_spec.parameter_names) == ["duplicates"]

    assert sorted(analysis_spec.column_names) == [
        "concatenated",
        "file1",
        "file2",
    ]

    assert sorted(analysis_spec.pipeline_names) == []
    assert sorted(analysis_spec.check_names) == []
    assert sorted(analysis_spec.switch_names) == []
    assert sorted(analysis_spec.subanalysis_names) == [
        "basic_sub",
        "outer_sub1",
        "outer_sub2",
    ]

    duplicates = analysis_spec.parameter("duplicates")
    assert duplicates.type is int
    assert duplicates.default == 4
    assert duplicates.salience == ps.recommended
    # Not sure why this is failing, not super critical at this point
    # assert common_duplicates.defined_in == (ConcatWithSubanalyses,)

    file1 = analysis_spec.column_spec("file1")
    assert file1.type is Zip[Text]
    assert file1.row_frequency == Samples.sample
    assert file1.salience == cs.primary
    # assert file1.defined_in == (Concat,)
    assert file1.mapped_from == ("basic_sub", "file1")

    file2 = analysis_spec.column_spec("file2")
    assert file2.type is Zip[Text]
    assert file2.row_frequency == Samples.sample
    assert file2.salience == cs.primary
    # assert file2.defined_in == (Concat,)
    assert file2.mapped_from == ("outer_sub2", "file1")

    basic_sub = analysis_spec.subanalysis_spec("basic_sub")
    assert basic_sub.name == "basic_sub"
    assert basic_sub.type is ExtendedConcat
    assert basic_sub.mappings == (
        ("doubly_concatenated", "concatenated"),
        ("duplicates", "duplicates"),
        ("file1", "file1"),
        ("file2", "file2"),
    )

    outer_sub1 = analysis_spec.subanalysis_spec("outer_sub1")
    assert outer_sub1.name == "outer_sub1"
    assert outer_sub1.type is ConcatWithSubanalyses
    assert outer_sub1.mappings == (
        ("concatenated", "concatenated"),
        ("file1", "file1"),
        ("file2", "file2"),
    )

    outer_sub2 = analysis_spec.subanalysis_spec("outer_sub2")
    assert outer_sub2.name == "outer_sub2"
    assert outer_sub2.type is ConcatWithSubanalyses
    assert outer_sub2.mappings == (
        ("common_duplicates", "duplicates"),
        ("file1", "file2"),
        ("file2", "file1"),
    )

    # Initialise class
    analysis = ConcatWithNestedSubanalyses(
        dataset=test_partial_numeric_dataset,
        file1="a_column",
        file2="another_column",
        duplicates=3,
        outer_sub1={"common_duplicates": 7, "sub2": {"multiplier": 1000}},
        outer_sub2={"sub2": {"multiplier": 10000}},
    )
    assert analysis.file1 == test_partial_numeric_dataset["a_column"]
    assert analysis.file2 == test_partial_numeric_dataset["another_column"]
    assert analysis.duplicates == 3
    assert analysis.basic_sub.file1 == test_partial_numeric_dataset["a_column"]
    assert analysis.basic_sub.file2 == test_partial_numeric_dataset["another_column"]
    assert analysis.basic_sub.duplicates == 3

    assert analysis.outer_sub1.file1 == test_partial_numeric_dataset["a_column"]
    assert analysis.outer_sub1.file2 == test_partial_numeric_dataset["another_column"]
    assert analysis.outer_sub1.common_duplicates == 7
    assert analysis.outer_sub1.concatenated is None
    assert analysis.outer_sub1.sub1.file1 == test_partial_numeric_dataset["a_column"]
    assert (
        analysis.outer_sub1.sub1.file2 == test_partial_numeric_dataset["another_column"]
    )
    assert analysis.outer_sub1.sub2.duplicates == 7
    assert analysis.outer_sub1.sub1.concatenated is None
    assert analysis.outer_sub1.sub1.doubly_concatenated is None
    assert analysis.outer_sub1.sub2.file1 == test_partial_numeric_dataset["a_column"]
    assert (
        analysis.outer_sub1.sub2.file2 == test_partial_numeric_dataset["another_column"]
    )
    assert analysis.outer_sub1.sub2.duplicates == 7
    assert analysis.outer_sub1.sub2.multiplier == 1000
    assert analysis.outer_sub1.sub2.concatenated is None
    assert analysis.outer_sub1.sub2.multiplied is None

    assert analysis.outer_sub2.file1 == test_partial_numeric_dataset["another_column"]
    assert analysis.outer_sub2.file2 == test_partial_numeric_dataset["a_column"]
    assert analysis.outer_sub2.common_duplicates == 3
    assert analysis.outer_sub2.concatenated is None
    assert (
        analysis.outer_sub2.sub1.file1 == test_partial_numeric_dataset["another_column"]
    )
    assert analysis.outer_sub2.sub1.file2 == test_partial_numeric_dataset["a_column"]
    assert analysis.outer_sub2.sub2.duplicates == 3
    assert analysis.outer_sub2.sub1.concatenated is None
    assert analysis.outer_sub2.sub1.doubly_concatenated is None
    assert (
        analysis.outer_sub2.sub2.file1 == test_partial_numeric_dataset["another_column"]
    )
    assert analysis.outer_sub2.sub2.file2 == test_partial_numeric_dataset["a_column"]
    assert analysis.outer_sub2.sub2.duplicates == 3
    assert analysis.outer_sub2.sub2.multiplier == 10000
    assert analysis.outer_sub2.sub2.concatenated is None
    assert analysis.outer_sub2.sub2.multiplied is None

    wf = pydra.Workflow(
        name="test_analysis",
        input_spec=["file1", "file2"],
        file1=test_numeric_file1,
        file2=test_numeric_file2,
    )
    concatenated = analysis.outer_sub2.sub2.concat_pipeline(
        wf,
        file1=wf.lzin.file1,
        file2=wf.lzin.file2,
        duplicates=analysis.outer_sub2.sub2.duplicates,
    )
    concat_and_multiplied = analysis.outer_sub2.sub2.multiply_pipeline(
        wf, concatenated=concatenated, multiplier=analysis.outer_sub2.sub2.multiplier
    )
    doubly_concatenated = analysis.outer_sub2.sub1.doubly_concat_pipeline(
        wf,
        concatenated=concatenated,
        file3=concat_and_multiplied,
        duplicates=analysis.duplicates,
    )
    wf.set_output(("doubly_concatenated", doubly_concatenated))
    result = wf(plugin="serial")
    assert get_contents(result.output.doubly_concatenated) == [
        "1",
        "2",
        "1",
        "2",
        "1",
        "2",
        "10000.0",
        "20000.0",
        "10000.0",
        "20000.0",
        "10000.0",
        "20000.0",
        "1",
        "2",
        "1",
        "2",
        "1",
        "2",
        "10000.0",
        "20000.0",
        "10000.0",
        "20000.0",
        "10000.0",
        "20000.0",
        "1",
        "2",
        "1",
        "2",
        "1",
        "2",
        "10000.0",
        "20000.0",
        "10000.0",
        "20000.0",
        "10000.0",
        "20000.0",
    ]


def test_reserved_name_errors():

    with pytest.raises(ArcanaDesignError):

        @analysis(Samples)
        class A:
            dataset: Text = column("a reserved attribute")

    with pytest.raises(ArcanaDesignError):

        @analysis(Samples)
        class B:
            menu: Text = column("another reserved attribute")

    with pytest.raises(ArcanaDesignError):

        @analysis(Samples)
        class C:
            stack: Text = column("yet another reserved attribute")


def test_change_of_type_errors():
    @analysis(Samples)
    class A:
        x: Text = column("a reserved attribute", salience=cs.primary)
        y: Text = column("another column")

        @pipeline(y)
        def a_pipeline(self, wf, x: Text):
            wf.add(identity_file(name="identity", in_file=x))
            return wf.identity.lzout.out_file

    with pytest.raises(ArcanaDesignError) as e:

        @analysis(Samples)
        class B(A):
            x: Zip[Text] = inherit()

    assert "Cannot change datatype" in e.value.msg


def test_multiple_pipeline_builder_errors():

    with pytest.raises(ArcanaDesignError) as e:

        @analysis(Samples)
        class A:
            x: Text = column("a column", salience=cs.primary)
            y: Text = column("another column")

            @pipeline(y)
            def a_pipeline(self, wf, x: Text):
                wf.add(identity_file(name="identity", in_file=x))
                return wf.identity.lzout.out_file

            @pipeline(y)
            def another_pipeline(self, wf, x: Text):
                wf.add(identity_file(name="identity", in_file=x))
                return wf.identity.lzout.out_file

    assert "Multiple pipelines provide outputs for 'y'" in e.value.msg


def test_unconnected_column_errors():

    # Should work
    @analysis(Samples)
    class A:
        x: Text = column("a column", salience=cs.primary)
        y: Text = column("another column")

        @pipeline(y)
        def a_pipeline(self, wf, x: Text):
            wf.add(identity_file(name="identity", in_file=x))
            return wf.identity.lzout.out_file

    with pytest.raises(ArcanaDesignError) as e:

        @analysis(Samples)
        class B:
            x: Text = column("a column", salience=cs.primary)

    assert "'x' is neither an input nor output to any pipeline" in e.value.msg

    with pytest.raises(ArcanaDesignError) as e:

        @analysis(Samples)
        class C:
            x: Text = column("a column")
            y: Text = column("another column")

            @pipeline(y)
            def a_pipeline(self, wf, x: Text):
                wf.add(identity_file(name="identity", in_file=x))
                return wf.identity.lzout.out_file

    assert (
        "'x' is not generated by any pipeline yet its salience is not specified as 'raw' or 'primary'"
        in e.value.msg
    )


def test_unknown_column_errors():

    with pytest.raises(ArcanaDesignError) as e:

        @analysis(Samples)
        class A:
            k: int = parameter("a parameter", default=1)

            @pipeline(k)
            def a_pipeline(self, wf):
                pass

    assert "'a_pipeline' pipeline outputs to unknown columns" in e.value.msg


def test_automagic_arg_errors():

    with pytest.raises(ArcanaDesignError) as e:

        @analysis(Samples)
        class A:
            y: Text = column("another column")

            @pipeline(y)
            def a_pipeline(self, wf, x: Text):
                wf.add(identity_file(name="identity", in_file=x))
                return wf.identity.lzout.out_file

    assert "Unrecognised argument 'x'" in e.value.msg


def test_inherit_errors():
    @analysis(Samples)
    class A:
        x: Text = column("a reserved attribute", salience=cs.primary)
        y: Text = column("another column")

        @pipeline(y)
        def a_pipeline(self, wf, x: Text):
            wf.add(identity_file(name="identity", in_file=x))
            return wf.identity.lzout.out_file

    with pytest.raises(AttributeError) as e:

        @analysis(Samples)
        class B(A):
            z: Text = inherit()

    assert "have no attribute named 'z' to inherit" in str(e.value)

    # @analysis(Samples)
    # class C(A):
    #     pass

    # with pytest.raises(ArcanaDesignError) as e:

    #     @analysis(Samples)
    #     class D(C):
    #         x: Text = inherit()

    # assert (
    #     "'x' must inherit from a column that is explicitly defined in the base class it references"
    #     in e.value.msg
    # )


def test_defined_in():
    @analysis(Samples)
    class A:
        x: Text = column("a reserved attribute", salience=cs.primary)
        y: Text = column("another column")

        @pipeline(y)
        def a_pipeline(self, wf, x: Text):
            wf.add(identity_file(name="identity", in_file=x))
            return wf.identity.lzout.out_file

    @analysis(Samples)
    class E(A):

        y: Text = inherit(salience=cs.publication)
        z: Text = column("yet another column", salience=cs.primary)

        @pipeline(z)
        def another_pipeline(self, wf, y: Text) -> Text:

            wf.add(identity_file(name="identity", in_file=y))

            return wf.identity.lzout.out_file

    @analysis(Samples)
    class F(E):

        x: Text = inherit(ref=A.x)
        z: Text = inherit(ref=E.z)

        @pipeline(z)
        def another_pipeline(self, wf, x: Text) -> Text:

            wf.add(identity_file(name="identity", in_file=x))

            return wf.identity.lzout.out_file

    assert attrs.fields(F).x.metadata[ARCANA_SPEC].defined_in == (A,)


def test_pipeline_overrides():
    @analysis(Samples)
    class A:
        x: Text = column("a reserved attribute", salience=cs.primary)
        y: Text = column("another column")

        @pipeline(y)
        def a_pipeline(self, wf, x: Text):
            wf.add(identity_file(name="identity", in_file=x))
            return wf.identity.lzout.out_file

    @analysis(Samples)
    class B(A):

        x: Text = inherit(ref=A.x)
        y: Text = inherit(ref=A.y)
        z: Text = column("yet another column")

        @pipeline(y, z)
        def a_pipeline(self, wf, x: Text):
            wf.add(identity_file(name="identity", in_file=x))
            return wf.identity.lzout.out_file, wf.identity.lzout.out_file

    with pytest.raises(ArcanaDesignError) as e:

        @analysis(Samples)
        class C(B):

            x: Text = inherit(ref=A.x)
            z: Text = inherit()

            @pipeline(z)
            def a_pipeline(self, wf, x: Text):
                wf.add(identity_file(name="identity", in_file=x))
                return wf.identity.lzout.out_file

    assert "['y'] outputs are missing from 'a_pipeline'" in e.value.msg


def test_parameter_bounds_validation(test_dataset):
    @analysis(Samples)
    class A:
        x: Text = column("a reserved attribute", salience=cs.primary)
        y: Text = column("another column")

        a: int = parameter("an int parameter", default=2, lower_bound=2)
        b: float = parameter("a float parameter", default=5, upper_bound=10)
        c: float = parameter(
            "a float parameter", default=10, lower_bound=2, upper_bound=20
        )
        required: str = parameter(
            "a required parameter", salience=ps.required, choices=["choice1", "choice2"]
        )

        @pipeline(y)
        def a_pipeline(self, wf, x: Text, a: int, b: float, required: str) -> Text:
            wf.add(identity_file(name="identity", in_file=x))
            return wf.identity.lzout.out_file

    with pytest.raises(ValueError) as e:
        A(dataset=test_dataset, x="a_column", a=1, required="choice1")

    assert "Value of 'a' (1) is not within the specified bounds" in str(e.value)

    with pytest.raises(ValueError) as e:
        A(dataset=test_dataset, x="file1", b=10.01, required="choice1")

    assert "Value of 'b' (10.01) is not within the specified bounds" in str(e.value)

    with pytest.raises(ValueError) as e:
        A(dataset=test_dataset, x="file1", c=-99, required="choice1")

    assert "Value of 'c' (-99) is not within the specified bounds" in str(e.value)

    with pytest.raises(ValueError) as e:
        A(
            dataset=test_dataset,
            x="file1",
            a=5,
            b=8.5,
        )

    assert "A value needs to be provided to required parameter 'required'" in str(
        e.value
    )

    with pytest.raises(ValueError) as e:
        A(dataset=test_dataset, x="file1", a=5, b=8.5, required="bad_choice")

    assert "'bad_choice' is not a valid value for 'required'" in str(e.value)

    with pytest.raises(ValueError) as e:

        @analysis(Samples)
        class B:
            a: int = parameter("bad default", default=-1, lower_bound=0)

    assert "is lower than lower bound" in str(e.value)

    with pytest.raises(ValueError) as e:

        @analysis(Samples)
        class C:
            a: int = parameter("bad default", default=100, upper_bound=99)

    assert "is higher than upper bound" in str(e.value)


def test_test_dataset(test_dataset):

    list(test_dataset["a_column"])



from __future__ import annotations
from operator import itemgetter
import pytest
from fileformats.text import Plain as PlainText
from arcana.core.exceptions import ArcanaUsageError
from arcana.stdlib import DirTree, Clinical
from arcana.testing.data.blueprint import TestDatasetBlueprint, FileSetEntryBlueprint
from arcana.testing.data.space import TestDataSpace


TEST_INCLUSIONS = {
    "all": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c0d0",
            "a0b0c0d1",
            "a0b0c0d2",
            "a0b0c0d3",
            "a0b0c1d0",
            "a0b0c1d1",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b0c2d0",
            "a0b0c2d1",
            "a0b0c2d2",
            "a0b0c2d3",
            "a0b1c0d0",
            "a0b1c0d1",
            "a0b1c0d2",
            "a0b1c0d3",
            "a0b1c1d0",
            "a0b1c1d1",
            "a0b1c1d2",
            "a0b1c1d3",
            "a0b1c2d0",
            "a0b1c2d1",
            "a0b1c2d2",
            "a0b1c2d3",
        ],
    ),
    "include": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            include={
                "c": ["c1", "c2"],
                "a": "a0",
                "d": ["d1", "d2", "d3"],
            },
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c1d1",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b0c2d1",
            "a0b0c2d2",
            "a0b0c2d3",
            "a0b1c1d1",
            "a0b1c1d2",
            "a0b1c1d3",
            "a0b1c2d1",
            "a0b1c2d2",
            "a0b1c2d3",
        ],
    ),
    "exclude": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            exclude={
                "c": ["c2"],
                "abcd": ["a0b0c1d1", "a0b1c1d2"],
            },
            id_patterns={"d": r"abcd::.*(d\d+)"},
        ),
        [
            "a0b0c0d0",
            "a0b0c0d1",
            "a0b0c0d2",
            "a0b0c0d3",
            "a0b0c1d0",
            "a0b0c1d2",
            "a0b0c1d3",
            "a0b1c0d0",
            "a0b1c0d1",
            "a0b1c0d2",
            "a0b1c0d3",
            "a0b1c1d0",
            "a0b1c1d1",
            "a0b1c1d3",
        ],
    ),
    "regex": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            id_patterns={"d": r"abcd::.*(d\d+)"},
            include={
                "abcd": r"a\d+b\dc(\d+)d\1",
            },
        ),
        [
            "a0b0c0d0",
            "a0b0c1d1",
            "a0b0c2d2",
            "a0b1c0d0",
            "a0b1c1d1",
            "a0b1c2d2",
        ],
    ),
}


@pytest.mark.parametrize("fixture", TEST_INCLUSIONS.items(), ids=itemgetter(0))
def test_dataset_inclusion(
    fixture: tuple[str, tuple[TestDatasetBlueprint, list[str]]], work_dir
):
    test_name, (blueprint, expected) = fixture
    dataset_path = work_dir / test_name
    dataset = blueprint.make_dataset(store=DirTree(), dataset_id=dataset_path)
    assert sorted(dataset.row_ids()) == expected


def test_include_exclude_fail1(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        space=TestDataSpace,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=PlainText, filenames=["file1.txt"]
            ),
        ],
        include={"session": r"a0.*d1.*"},
    )
    with pytest.raises(ArcanaUsageError, match="Unrecognised frequencies in 'include'"):
        blueprint.make_dataset(
            store=DirTree(), dataset_id=work_dir / "include-exclude-fail1"
        )


def test_include_exclude_fail2(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        space=TestDataSpace,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=PlainText, filenames=["file1.txt"]
            ),
        ],
        exclude={"bc": r"a0.*d1.*"},
    )
    with pytest.raises(
        ArcanaUsageError,
        match="only frequencies present in the dataset hierarchy are allowed",
    ):
        blueprint.make_dataset(
            store=DirTree(), dataset_id=work_dir / "include-exclude-fail2"
        )


def test_include_exclude_fail3(work_dir):

    blueprint = TestDatasetBlueprint(  # dataset name
        space=TestDataSpace,
        hierarchy=["a", "b", "c", "abcd"],
        dim_lengths=[1, 2, 3, 4],
        entries=[
            FileSetEntryBlueprint(
                path="file1", datatype=PlainText, filenames=["file1.txt"]
            ),
        ],
        include={"abcd": r"a**"},
    )
    with pytest.raises(ArcanaUsageError, match="valid regular expression"):
        blueprint.make_dataset(
            store=DirTree(), dataset_id=work_dir / "include-exclude-fail3"
        )


TEST_AUTO_IDS = {
    "d_dim": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["a", "b", "c", "abcd"],
            dim_lengths=[1, 2, 3, 4],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
        ),
        {
            "d": ["1", "2", "3", "4"],
        },
    ),
    "member": (
        TestDatasetBlueprint(  # dataset name
            space=Clinical,
            hierarchy=["group", "subject", "timepoint"],
            dim_lengths=[2, 2, 2],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
        ),
        {
            "member": ["1", "2"],
        },
    ),
    "double_increment": (
        TestDatasetBlueprint(  # dataset name
            space=TestDataSpace,
            hierarchy=["ab", "cd"],
            dim_lengths=[2, 2, 2, 2],
            entries=[
                FileSetEntryBlueprint(
                    path="file1", datatype=PlainText, filenames=["file1.txt"]
                ),
            ],
            id_patterns={
                "b": r"ab::a\d+(b\d+)",
                "c": r"cd::(c\d+)d\d+",
            },
        ),
        {
            "a": ["1", "2"],
            "b": ["b0", "b1"],
            "c": ["c0", "c1"],
            "d": ["1", "2"],
        },
    ),
}


@pytest.mark.parametrize("fixture", TEST_AUTO_IDS.items(), ids=itemgetter(0))
def test_auto_ids(work_dir, fixture):

    _, (blueprint, expected) = fixture

    dataset = blueprint.make_dataset(
        store=DirTree(), dataset_id=work_dir / "incrementing-ids"
    )

    for key, ids in expected.items():
        assert sorted(dataset.row_ids(key)) == ids



from functools import reduce
from operator import mul
from pathlib import Path
import pytest
from arcana.testing.data.blueprint import (
    TestDatasetBlueprint,
    FileSetEntryBlueprint as FileBP,
)
from arcana.testing.data import TestDataSpace
from fileformats.text import Plain as Text
from fileformats.testing import EncodedText
from arcana.core.data.set import Dataset
from arcana.stdlib import DirTree
from arcana.core.deploy.command.base import ContainerCommand
from arcana.core.exceptions import ArcanaDataMatchError


def test_command_execute(concatenate_task, saved_dataset, work_dir):
    # Get CLI name for dataset (i.e. file system path prepended by 'dirtree//')
    bp = saved_dataset.__annotations__["blueprint"]
    duplicates = 1

    command_spec = ContainerCommand(
        task="arcana.testing.tasks:" + concatenate_task.__name__,
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "source1",
                "datatype": "fileformats.text:Plain",
                "field": "in_file1",
                "help_string": "dummy",
            },
            {
                "name": "source2",
                "datatype": "fileformats.text:Plain",
                "field": "in_file2",
                "help_string": "dummy",
            },
        ],
        outputs=[
            {
                "name": "sink1",
                "datatype": "fileformats.text:Plain",
                "field": "out_file",
                "help_string": "dummy",
            }
        ],
        parameters=[
            {
                "name": "duplicates",
                "datatype": "int",
                "default": 2,
                "help_string": "dummy",
            }
        ],
    )
    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    command_spec.execute(
        dataset_locator=saved_dataset.locator,
        input_values=[
            ("source1", "file1"),
            ("source2", "file2"),
        ],
        output_values=[
            ("sink1", "concatenated"),
        ],
        parameter_values=[
            ("duplicates", str(duplicates)),
        ],
        raise_errors=True,
        plugin="serial",
        work_dir=str(work_dir),
        loglevel="debug",
        dataset_hierarchy=",".join(bp.hierarchy),
        pipeline_name="test_pipeline",
    )
    # Add source column to saved dataset
    sink = saved_dataset.add_sink("concatenated", Text)
    assert len(sink) == reduce(mul, bp.dim_lengths)
    fnames = ["file1.txt", "file2.txt"]
    if concatenate_task.__name__.endswith("reverse"):
        fnames = [f[::-1] for f in fnames]
    expected_contents = "\n".join(fnames * duplicates)
    for item in sink:
        with open(item) as f:
            contents = f.read()
        assert contents == expected_contents


def test_command_execute_fail(concatenate_task, saved_dataset, work_dir):
    # Get CLI name for dataset (i.e. file system path prepended by 'dirtree//')
    bp = saved_dataset.__annotations__["blueprint"]
    duplicates = 1

    command_spec = ContainerCommand(
        task="arcana.testing.tasks:" + concatenate_task.__name__,
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "source1",
                "datatype": "fileformats.text:Plain",
                "field": "in_file1",
                "help_string": "dummy",
            },
            {
                "name": "source2",
                "datatype": "fileformats.generic:Directory",
                "field": "in_file2",
                "help_string": "dummy",
            },
        ],
        outputs=[
            {
                "name": "sink1",
                "datatype": "fileformats.text:Plain",
                "field": "out_file",
                "help_string": "dummy",
            }
        ],
        parameters=[
            {
                "name": "duplicates",
                "datatype": "int",
                "default": 2,
                "help_string": "dummy",
            }
        ],
    )

    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    with pytest.raises(ArcanaDataMatchError):
        command_spec.execute(
            dataset_locator=saved_dataset.locator,
            input_values=[
                ("source1", "bad-file-path"),
                ("source2", "file2"),
            ],
            output_values=[
                ("sink1", "concatenated"),
            ],
            parameter_values=[
                ("duplicates", duplicates),
            ],
            raise_errors=True,
            plugin="serial",
            work_dir=str(work_dir),
            loglevel="debug",
            dataset_hierarchy=",".join(bp.hierarchy),
            pipeline_name="test_pipeline",
        )


def test_command_execute_on_row(cli_runner, work_dir):

    # Create test dataset consisting of a single row with a range of filenames
    # from 0 to 4
    filenumbers = list(range(5))
    bp = TestDatasetBlueprint(
        space=TestDataSpace,
        hierarchy=[
            "abcd"
        ],  # e.g. XNAT where session ID is unique in project but final layer is organised by timepoint
        dim_lengths=[1, 1, 1, 1],
        entries=[
            FileBP(path=str(i), datatype=Text, filenames=[f"{i}.txt"])
            for i in filenumbers
        ],
    )
    dataset_path = work_dir / "numbered_dataset"
    dataset = bp.make_dataset(DirTree(), dataset_path)
    dataset.save()

    def get_dataset_filenumbers():
        row = next(iter(dataset.rows()))
        return sorted(int(i.path.split(".")[0]) for i in row.entries)

    assert get_dataset_filenumbers() == filenumbers

    command_spec = ContainerCommand(
        task="arcana.testing.tasks:plus_10_to_filenumbers",
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "a_row",
                "datatype": "arcana.core.data.row:DataRow",
                "field": "filenumber_row",
                "help_string": "dummy",
            },
        ],
    )

    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    command_spec.execute(
        dataset_locator=dataset.locator,
        input_values=[
            ("a_row", ""),
        ],
        raise_errors=True,
        plugin="serial",
        work_dir=str(work_dir),
        loglevel="debug",
        dataset_hierarchy=",".join(bp.hierarchy),
        pipeline_name="test_pipeline",
    )

    assert get_dataset_filenumbers() == [i + 10 for i in filenumbers]


def test_command_execute_with_converter_args(saved_dataset: Dataset, work_dir: Path):
    """Test passing arguments to file format converter tasks via input/output
    "qualifiers", e.g. 'converter.shift=3' using the arcana-run-pipeline CLI
    tool (as used in the XNAT CS commands)
    """
    # Get CLI name for dataset (i.e. file system path prepended by 'dirtree//')
    bp = saved_dataset.__annotations__["blueprint"]
    # Start generating the arguments for the CLI
    # Add source to loaded dataset
    command_spec = ContainerCommand(
        task="arcana.testing.tasks:identity_file",
        row_frequency=bp.space.default(),
        inputs=[
            {
                "name": "source",
                "datatype": "fileformats.testing:EncodedText",
                "default_column": {"datatype": "fileformats.text:Plain"},
                "field": "in_file",
                "help_string": "dummy",
            },
        ],
        outputs=[
            {
                "name": "sink1",
                "datatype": "fileformats.testing:EncodedText",
                "field": "out",
                "help_string": "dummy",
            },
            {
                "name": "sink2",
                "datatype": "fileformats.testing:EncodedText",
                "default_column": {"datatype": "fileformats.text:Plain"},
                "field": "out",
                "help_string": "dummy",
            },
        ],
    )

    command_spec.execute(
        dataset_locator=saved_dataset.locator,
        input_values=[
            ("source", "file1 converter.shift=3"),
        ],
        output_values=[
            ("sink1", "encoded"),
            ("sink2", "decoded converter.shift=-3"),
        ],
        raise_errors=True,
        plugin="serial",
        work_dir=str(work_dir),
        loglevel="debug",
        dataset_hierarchy=",".join(bp.hierarchy),
        pipeline_name="test_pipeline",
    )
    # Add source column to saved dataset
    saved_dataset.add_sink("sink1", EncodedText, path="encoded@")
    saved_dataset.add_sink("sink2", Text, path="decoded@")
    unencoded_contents = "file1.txt"
    encoded_contents = (
        "iloh41w{w"  # 'file1.txt' characters shifted up by 3 in ASCII code
    )
    for row in saved_dataset.rows(frequency="abcd"):
        enc_cell = row.cell("sink1", allow_empty=False)
        dec_cell = row.cell("sink2", allow_empty=False)
        enc_item = enc_cell.item
        dec_item = dec_cell.item
        with open(enc_item) as f:
            enc_contents = f.read()
        with open(dec_item) as f:
            dec_contents = f.read()
        assert enc_contents == encoded_contents
        assert dec_contents == unencoded_contents
